import numpy as np
from Bvalcalc.core.helpers.calc_B_from_chunks import calc_B_from_chunks
from Bvalcalc.core.helpers.calc_B_precise_noninterfering import calc_B_precise_noninterfering
from Bvalcalc.core.calculateB import calculateB_hri

def calc_B_in_hri_region(quiet, chunk_idx, rec_rate_per_chunk, hri_r_threshold, lperchunk, chunk_size, chr_start, chr_size, num_chunks, gc_rate_per_chunk, precise_chunks, precise_blockstart, precise_blockend, pos_chunk, chunk_end, precise_region_start, precise_region_end, unlinked_B):
    if quiet: print(f"Chunk {chunk_idx}: r modifier = {rec_rate_per_chunk[chunk_idx]}, which is at or below 0.1 threshold. Calculating B'. HRI calculation is enabled with --hri")
    low_rec_chunk_ids = rec_rate_per_chunk < hri_r_threshold

    # extend left, then right til there's a chunk with recombination
    interference_region_start_idx, interference_region_end_idx = chunk_idx, chunk_idx
    while interference_region_start_idx > 0 and low_rec_chunk_ids[interference_region_start_idx - 1]:
        interference_region_start_idx -= 1
    while interference_region_end_idx < low_rec_chunk_ids.size - 1 and low_rec_chunk_ids[interference_region_end_idx + 1]:
        interference_region_end_idx += 1
    total_interfering_L = lperchunk[interference_region_start_idx : interference_region_end_idx + 1].sum()## get combined 
    # if quiet: print(f"Contiguous low‐rec run: chunks {interference_region_start_idx}–{interference_region_end_idx}, interfering sites L = {total_interfering_L:.1f}")

    local_interference_indices = np.arange(interference_region_start_idx, interference_region_end_idx + 1) # Indexes for chunks in local interfering region

    B_from_distant_chunks = calc_B_from_chunks( # Re-compute B from distant chunks in non-precise region, exluding local interfering region
        chunk_idx, chunk_size, chr_start, chr_size, num_chunks, 
        precise_chunks, lperchunk, rec_rate_per_chunk, gc_rate_per_chunk, local_interference_indices)
    # print(f"Chunk {chunk_idx}: B_from_distant_chunks, excluding local interference region", B_from_distant_chunks)
    
    B_noninterfering_in_precise_region = calc_B_precise_noninterfering(precise_blockstart, precise_blockend, pos_chunk,
                                                                    chr_start, chunk_end, chunk_size, chr_size, precise_region_start, precise_region_end,local_interference_indices, chunk_idx, 
                                                                    rec_rate_per_chunk, gc_rate_per_chunk)

    U_lengths_in_low_rec_chunks = lperchunk[low_rec_chunk_ids]

    combined_prior_B = B_from_distant_chunks * B_noninterfering_in_precise_region[0] * unlinked_B

    interference_Bvals_per_chunk = calculateB_hri(
        distant_B=combined_prior_B,
        interfering_L=U_lengths_in_low_rec_chunks
    )

    return interference_Bvals_per_chunk


import os
import sys

# keep your list of valid species here - updated for new naming convention
SPECIES = [
    'selfing',  # Special case - no element type
    'aratha_cds', 'aratha_phastcons',  # Arabidopsis thaliana
    'dromel_cds', 'dromel_utr', 'dromel_phastcons',  # Drosophila melanogaster
    'homsap_cds',  # Homo sapiens
]

def check_generate_params_args(argv=None):
    """
    If '--generate_params' appears with no species or next flag,
    print a concise error and exit.
    """
    if argv is None:
        argv = sys.argv
    if '--generate_params' in argv:
        idx = argv.index('--generate_params')
        if idx == len(argv) - 1 or argv[idx+1].startswith('-'):
            print(f"Provide name of default template as an argument: {' '.join(SPECIES)}")
            sys.exit(1)

def generateParams(species, folder='.'):
    # Convert to lowercase for case-insensitive matching
    species_lower = species.lower()
    
    # Handle special case for selfing
    if species_lower == 'selfing':
        tpl_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                '..', 'templates',
                'SelfingParams.py'
            )
        )
        dest_name = 'SelfingParams.py'
    else:
        # Handle new naming convention: GenSpe_ElementType_Params.py
        # Map from command format to file format
        species_mapping = {
            'aratha_cds': 'AraTha_Cds_Params.py',
            'aratha_phastcons': 'AraTha_Phastcons_Params.py',
            'dromel_cds': 'DroMel_Cds_Params.py',
            'dromel_utr': 'DroMel_Utr_Params.py',
            'dromel_phastcons': 'DroMel_Phastcons_Params.py',
            'homsap_cds': 'HomSap_Cds_Params.py',
        }
        
        if species_lower not in species_mapping:
            print(f"Error: Unknown species '{species}'.")
            print(f"Available options: {', '.join(SPECIES)}")
            print("Note: Species names are case-insensitive (e.g., 'ARATHA_CDS' works the same as 'aratha_cds')")
            sys.exit(1)
        
        filename = species_mapping[species_lower]
        tpl_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                '..', 'templates',
                filename
            )
        )
        dest_name = filename
    
    if not os.path.isfile(tpl_path):
        raise FileNotFoundError(f"Template for '{species}' not found at {tpl_path}")

    # Read the template
    with open(tpl_path, 'r') as tpl_file:
        content = tpl_file.read()
    print(f"Loaded template from:   {tpl_path}")

    # Ensure target folder exists
    os.makedirs(folder, exist_ok=True)

    # Write into <folder>/<dest_name>
    dest_path = os.path.join(folder, dest_name)
    with open(dest_path, 'w') as out_file:
        out_file.write(content)

    print(f"Wrote parameters to:     {dest_path}")
    print("Note that these are example parameters, please tailor to your population and analysis")


import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.ticker as ticker
import numpy as np
from matplotlib.collections import LineCollection
from matplotlib import gridspec  # for rec rate strip

def plotB(b_values_input, caller, output_path, quiet, gene_ranges=None, neutral_only=False, rec_rates=None, chunk_size=None):
    if not quiet:
        print('====== P L O T T I N G . . . =======================')

    # Configure fonts and styles
    mpl.rcParams['font.family'] = ['Helvetica', 'DejaVu Sans', 'Arial']
    if 'seaborn-v0_8-whitegrid' in plt.style.available:
        plt.style.use('seaborn-v0_8-whitegrid')
    else:
        print("'seaborn-whitegrid' style not found, using 'ggplot' as fallback.")
        plt.style.use('ggplot')
        mpl.rcParams['axes.facecolor'] = 'white'
        mpl.rcParams['figure.facecolor'] = 'white'
        mpl.rcParams['grid.color'] = 'grey'
        mpl.rcParams['grid.linestyle'] = '--'
        mpl.rcParams['grid.linewidth'] = 0.5
    mpl.rcParams['axes.edgecolor'] = 'black'

    # Create figure and axis based on rec_rates
    if rec_rates is None:
        fig, ax = plt.subplots(figsize=(10, 6))
    else:
        height_ratios = [10, 0.5]
        fig = plt.figure(figsize=(10, 6))
        gs = gridspec.GridSpec(nrows=2, ncols=1, height_ratios=height_ratios)
        ax = fig.add_subplot(gs[0])

    # Main plotting logic
    if caller == "chromosome":
        positions = b_values_input['Start']
        b_vals = b_values_input['B']
        chrom = b_values_input['Chromosome'][0] if 'Chromosome' in b_values_input.dtype.names else 'unknown'

        if neutral_only:
            conserved = b_values_input['Conserved']
            is_bytes = conserved.dtype.kind == 'S'
            neutral_mask = conserved == b'N' if is_bytes else conserved == 'N'

            x = positions[neutral_mask]
            y = b_vals[neutral_mask]
            sort_idx = np.argsort(x)
            x = x[sort_idx]
            y = y[sort_idx]

            diffs = np.diff(x)
            split_indices = np.where(diffs != 1)[0] + 1
            x_segments = np.split(x, split_indices)
            y_segments = np.split(y, split_indices)

            print(f"Plotting {len(x)} neutral positions in {len(x_segments)} segments.")
            for xs, ys in zip(x_segments, y_segments):
                if len(xs) > 1:
                    ax.plot(xs, ys, color='blue', lw=1.5, alpha=0.8)

            if len(x) > 0:
                ax.set_xlim(x.min() - 1, x.max())
        else:
            x = positions
            y = b_vals
            max_points = 10000
            if len(x) > max_points:
                idx = np.linspace(0, len(x) - 1, max_points).astype(int)
                x = x[idx]
                y = y[idx]
            ax.plot(x, y, color='blue', lw=1.5, alpha=0.8)
            ax.set_xlim(x.min() - 1, x.max())

    elif caller == "gene":
        x = b_values_input[:, 0]
        y = b_values_input[:, 1]
        ax.plot(x, y, color='blue', lw=1.5, alpha=0.8)
        ax.set_xlim(x.min() - 1, x.max())

    # Labels and title
    ax.set_ylabel('Expected diversity relative to neutral evolution (B)', fontsize=13)
    if caller == "chromosome":
        ax.set_title(f'B for chromosome {chrom} ({positions.min()}–{positions.max()} bp)', fontsize=15, fontweight='bold')
        if rec_rates is not None:
            ax.set_xlabel('Chromosomal position (bp)', fontsize=13, labelpad=40)
        else:
            ax.set_xlabel('Chromosomal position (bp)', fontsize=13)
    else:
        ax.set_xlabel('Distance from single selected element of size', fontsize=13)
        ax.set_title('B recovery from single element', fontsize=15, fontweight='bold')

    ax.tick_params(axis='both', which='major', labelsize=10)

    # Gene-range bars
    if gene_ranges is not None and len(gene_ranges) > 0:
        if caller == "chromosome":
            ymin, ymax = ax.get_ylim()
            bar_y = ymin - (ymax - ymin) * 0.05

            # compute the current x-range
            xlim = ax.get_xlim()
            xrange = xlim[1] - xlim[0]
            fig_width_in_pixels = fig.bbox.width
            min_width = xrange / fig_width_in_pixels

            # convert start/end to int for plotting, ensuring min width
            segments = []
            for _, start, end in gene_ranges:
                s = float(start)
                e = float(end)
                if (e - s) < min_width:
                    e = s + min_width
                segments.append(((s, bar_y), (e, bar_y)))

            ax.set_ylim(bar_y, ymax)
            ax.add_collection(LineCollection(segments, colors='black', linewidths=30))

            if not neutral_only:
                gene_mask = np.zeros_like(x, dtype=bool)
                for _, start, end in gene_ranges:
                    start = int(start)
                    end = int(end)
                    gene_mask |= (x >= start) & (x <= end)
                idx = np.where(gene_mask)[0]
                splits = np.where(np.diff(idx) != 1)[0] + 1
                for seg in np.split(idx, splits):
                    if len(seg) > 1:
                        coords = np.column_stack((x[seg], y[seg]))
                        ax.add_collection(LineCollection([coords], colors='black', linewidths=1.5))
        else:
            # For caller != chromosome, keep original behavior (no min width enforcement)
            ymin, ymax = ax.get_ylim()
            bar_y = ymin - (ymax - ymin) * 0.05
            ax.set_ylim(bar_y, ymax)
            segments = [((int(start), bar_y), (int(end), bar_y)) for _, start, end in gene_ranges]
            ax.add_collection(LineCollection(segments, colors='black', linewidths=30))

    # X-axis formatting
    ax.xaxis.set_major_formatter(
        ticker.FuncFormatter(lambda value, pos:
            f"{int(value)} bp" if value < 1e3 else (
                f"{value/1e6:.2f} Mb" if value >= 1e6 else f"{int(value/1e3)} kb"
            )
        )
    )

    from matplotlib.colors import LinearSegmentedColormap
    magenta_map = LinearSegmentedColormap.from_list("custom_magenta", ["white", "#C54B8C"])
    
    # Recombination rate strip
    if rec_rates is not None and caller == "chromosome":
        ax_rec = fig.add_subplot(gs[1], sharex=ax)
        ax_rec.set_yticks([])
        ax_rec.tick_params(axis='x', which='major', labelsize=9)
        rec_img = np.expand_dims(rec_rates, axis=0)
        min_pos = positions.min()
        extent = [min_pos, min_pos + len(rec_rates) * chunk_size, 0, 1]
        ax_rec.imshow(rec_img, aspect='auto', extent=extent, cmap=magenta_map, origin='lower', zorder=2, vmin=0, vmax=np.max(rec_rates))
        ax_rec.set_frame_on(False)
        plt.setp(ax.get_xticklabels(), visible=False)

    # Final layout and save
    if rec_rates is None:
        try:
            plt.savefig(output_path, dpi=300)
        except Exception as e:           # you could catch TimeoutError specifically if you prefer
            import sys
            print(f"Warning: could not save plot to {output_path!r}: {e}", file=sys.stderr)
    else:
        fig.subplots_adjust(hspace=0.01, bottom=0.12)

    plt.savefig(output_path, dpi=300)
    print(f"Plot saved to {output_path}")



"""
Core calculation modules for bvalcalc.
"""
from .genomeBcalc import genomeBcalc
from .regionBcalc import regionBcalc
from .geneBcalc import geneBcalc
from .siteBcalc import siteBcalc
from .plotB import plotB
from .calculateB import calculateB_linear, calculateB_recmap, calculateB_unlinked

__all__ = [
    "genomeBcalc",
    "regionBcalc",
    "geneBcalc",
    "siteBcalc",
    "plotB",
    "calculateB_linear", 
    "calculateB_recmap", 
    "calculateB_unlinked"
]



import numpy as np
import csv
from Bvalcalc.utils.load_vcf import load_vcf
from Bvalcalc.utils.load_Bmap import load_Bmap

def positionsBstats(args, Bmap_path):

    if not args.quiet: print(f"= B-map utilities = = =")
    # Load VCF and B-map
    vcf_chroms, vcf_pos = load_vcf(args.positions)
    print(f"loading B-map: {Bmap_path}")
    bmap_chroms, bmap_pos, b_values = load_Bmap(file_path=Bmap_path)

    # Header for retrieval
    if not args.quiet:
        print("====== R E T R I E V I N G === B - V A L U E S =====")

    # Identify unique chromosomes
    bmap_unique = np.unique(bmap_chroms)
    vcf_unique  = np.unique(vcf_chroms)

    # Report mismatches
    missing_bmap = set(vcf_unique.astype(str)) - set(bmap_unique.astype(str))
    missing_vcf  = set(bmap_unique.astype(str)) - set(vcf_unique.astype(str))
    if missing_bmap:
        print("WARNING: Chromosomes in VCF/CSV but not B-map: " +
              ", ".join(sorted(missing_bmap)))
    if missing_vcf:
        print("WARNING: Chromosomes in B-map but not VCF/CSV: " +
              ", ".join(sorted(missing_vcf)))

    # Filter out VCF entries not in B-map
    if missing_bmap:
        keep_mask = np.isin(vcf_chroms, bmap_unique)
        vcf_chroms, vcf_pos = vcf_chroms[keep_mask], vcf_pos[keep_mask]
        vcf_unique = np.unique(vcf_chroms)

    # Prepare flat arrays
    n = vcf_pos.size
    flat_b     = np.empty(n, dtype=np.float64)
    flat_chrom = np.empty(n, dtype='<U20')

    # Prepare output file if requested
    writer = None
    if args.out:
        out_f = open(args.out, 'w', newline='')
        if not args.bcftools_format:
            # standard CSV with header
            writer = csv.writer(out_f)
            writer.writerow(['chromosome', 'position', 'B'])
        # else: plain "chromosome:position" lines, no header

    # Map B-values per chromosome
    for chrom in vcf_unique:
        mask_v  = (vcf_chroms == chrom)
        pos_chr = vcf_pos[mask_v]
        mask_b  = (bmap_chroms == chrom)
        starts, vals = bmap_pos[mask_b], b_values[mask_b]

        # Overflow warning
        max_start = int(starts.max())
        above_count_chr = int(np.count_nonzero(pos_chr > max_start))
        if above_count_chr > 0:
            print(f"WARNING: {above_count_chr} VCF/CSV positions in {chrom} "
                  f"are above max B-map start ({max_start})")

        # Assign B-values
        idx = np.searchsorted(starts, pos_chr, side='right') - 1
        idx[idx < 0] = 0
        flat_b[mask_v]     = vals[idx]
        flat_chrom[mask_v] = chrom

        # Write out
        if writer:
            for p, b in zip(pos_chr, vals[idx]):
                if (args.out_minimum is None or b >= args.out_minimum) and \
                   (args.out_maximum is None or b <= args.out_maximum):
                    writer.writerow([chrom, p, b])
        elif args.bcftools_format and args.out:
            for p, b in zip(pos_chr, vals[idx]):
                if (args.out_minimum is None or b >= args.out_minimum) and \
                   (args.out_maximum is None or b <= args.out_maximum):
                    out_f.write(f"{chrom}:{p}\n")

    # Summary stats
    filter_mask = np.ones(n, dtype=bool)
    if args.out_minimum is not None:
        filter_mask &= (flat_b >= args.out_minimum)
    if args.out_maximum is not None:
        filter_mask &= (flat_b <= args.out_maximum)

    filtered_b     = flat_b[filter_mask]
    filtered_chrom = flat_chrom[filter_mask]
    filtered_pos   = vcf_pos[filter_mask]
    n_filtered     = filtered_b.size

    if not args.quiet:
        print("====== R E S U L T S ====== S U M M A R Y ==========")
    if n_filtered > 0:
        mean_B    = float(filtered_b.mean())
        idx_max   = int(filtered_b.argmax())
        idx_min   = int(filtered_b.argmin())
        max_B     = float(filtered_b[idx_max])
        min_B     = float(filtered_b[idx_min])
        pos_max   = int(filtered_pos[idx_max])
        pos_min   = int(filtered_pos[idx_min])
        chrom_max = filtered_chrom[idx_max]
        chrom_min = filtered_chrom[idx_min]

        if not args.quiet:
            print(f"Mean B across filtered sites: {mean_B:.6f}")
            print(f"Max B across filtered sites: {max_B:.6f} at {chrom_max}:{pos_max}")
            print(f"Min B across filtered sites: {min_B:.6f} at {chrom_min}:{pos_min}")
    else:
        if not args.quiet:
            print("No B-values to summarize after applying thresholds.")

    # Close file if open
    if args.out:
        out_f.close()
        if not args.quiet:
            print(
                f"Wrote CSV to {args.out}"
                if not args.bcftools_format
                else f"Wrote regions to {args.out}"
            )
    else:
        if not args.quiet:
            print("Skipping save (use --out to write CSV or regions file)")

    return filtered_b, filtered_chrom



import numpy as np

def extend_hri_regions_correction(b_values, rec_rate_per_chunk, chunk_size, chr_start, calc_start, calc_end, hri_r_threshold = 0.1):

    ## First need to export properly from process_single_chunk.py

    # print(f"Extending HRI regions", b_values, rec_rate_per_chunk)

    low_rec_chunk_ids = rec_rate_per_chunk < hri_r_threshold

    mask = low_rec_chunk_ids
    if not mask.any():
        return b_values

    # Inclusive starts/ends of each True-run (contiguous HRI regions)
    interference_region_starts_idx = np.where(mask & np.r_[True, ~mask[:-1]])[0]
    interference_region_ends_idx   = np.where(mask & np.r_[~mask[1:], True])[0]

    # print(interference_region_starts_idx, interference_region_ends_idx)

    base_chunk_idx = (calc_start - chr_start) // chunk_size

    # Get start and end positions of interference regions
    abs_start_chunks = base_chunk_idx + interference_region_starts_idx
    abs_end_chunks   = base_chunk_idx + interference_region_ends_idx + 1  # +1 for end-exclusive

    interference_region_start_pos = chr_start + abs_start_chunks * chunk_size
    interference_region_end_pos = np.minimum(calc_end,
        chr_start + abs_end_chunks * chunk_size - 1)

    # Filter to only include interference regions that overlap with the calculation region
    valid_mask = (interference_region_end_pos >= calc_start) & (interference_region_start_pos <= calc_end)
    if not valid_mask.any():
        return b_values

    interference_region_start_pos = interference_region_start_pos[valid_mask]
    interference_region_end_pos = interference_region_end_pos[valid_mask]

    # Clip positions to be within the calculation region
    interference_region_start_pos = np.maximum(interference_region_start_pos, calc_start)
    interference_region_end_pos = np.minimum(interference_region_end_pos, calc_end)

    B_in_interference_regions = b_values[interference_region_start_pos - calc_start]

    # Debug peek
    # print("interference_region_start_pos:", interference_region_start_pos)
    # print("interference_region_end_pos  :", interference_region_end_pos)
    # print("b_values  :", B_in_interference_regions)

    left_extended  = np.zeros(len(interference_region_start_pos), dtype=int)
    right_extended = np.zeros(len(interference_region_end_pos), dtype=int)

    n = b_values.shape[0]


    # ---- Backward extension from each region start ----
    # For region i, extend left from start_pos[i] - 1 while region B is higher
    # than the current b_values. Stop at calc_start or just after the previous region.
    for i in range(len(interference_region_start_pos)):
        b_inside = B_in_interference_regions[i]

        # ---- backward extension ----
        idx_left = (interference_region_start_pos[i] - calc_start) - 1
        if i == 0:
            stop_left_rel = 0
        else:
            stop_left_abs = interference_region_end_pos[i - 1] + 1
            stop_left_rel = max(0, stop_left_abs - calc_start)

        while idx_left >= stop_left_rel:
            if b_inside > b_values[idx_left]:
                b_values[idx_left] = b_inside
                left_extended[i] += 1
                # NEW: stop if extended beyond chunk_size
                if left_extended[i] > chunk_size:
                    break
                idx_left -= 1
            else:
                break

        # ---- forward extension ----
        idx_right = (interference_region_end_pos[i] - calc_start) + 1
        if i == len(interference_region_start_pos) - 1:
            stop_right_rel = n - 1
        else:
            stop_right_abs = interference_region_start_pos[i + 1] - 1
            stop_right_rel = min(n - 1, stop_right_abs - calc_start)

        while idx_right <= stop_right_rel:
            if b_inside > b_values[idx_right]:
                b_values[idx_right] = b_inside
                right_extended[i] += 1
                # NEW: stop if extended beyond chunk_size
                if right_extended[i] > chunk_size:
                    break
                idx_right += 1
            else:
                break

    # ---- warnings ----
    for i in range(len(interference_region_start_pos)):
        warn_left  = left_extended[i]  > chunk_size
        warn_right = right_extended[i] > chunk_size
        if warn_left or warn_right:
            sides = []
            if warn_left:
                sides.append(f"left={left_extended[i]}")
            if warn_right:
                sides.append(f"right={right_extended[i]}")
            print(
                f"WARNING: HRI extension exceeded chunk_size at region {i} "
                f"[{interference_region_start_pos[i]}-{interference_region_end_pos[i]}]; "
                f"{', '.join(sides)} bases > chunk_size ({chunk_size}). "
                f"This suggests B is very low and HRI is too pervasive for B to be calculated effectively in this region."
            )

        # NEW: compute final (start,end) positions after extension, inclusive
    extended_starts = np.maximum(interference_region_start_pos - left_extended, calc_start)
    extended_ends   = np.minimum(interference_region_end_pos + right_extended, calc_end)
    # extended_regions_pos = np.stack([extended_starts, extended_ends], axis=1).astype(int)

    print(f"Extended HRI regions to {extended_starts, extended_ends}")

    return extended_starts, extended_ends


import csv
import numpy as np
import sys

def load_rec_map(rec_map, calc_start, calc_end, chunk_size, chromosome):
    """
    Processes the recombination map file (no header) and returns
    the average recombination rate per chunk, weighted by the proportion of each 
    chunk that falls within a given recombination interval.
    
    The file must be headerless, with each row having exactly three columns:
    chromosome (str), start (int), and rate (float). Only rows whose first
    column equals the provided `chromosome` are used; all others are skipped.
    
    Parameters:
        rec_map (str): Path to the recombination map file.
        calc_start (int): Starting position of the chromosome.
        calc_end (int): Ending position of the chromosome.
        chunk_size (int): Size of each chunk in base pairs.
        chromosome (str): Chromosome name to filter rows by.
    
    Returns:
        np.ndarray: Average recombination rate for each chunk.
    """
    rec_map_data = []
    with open(rec_map, newline='') as csvfile:
        reader = csv.reader(csvfile)
        for row in reader:
            # must have exactly [str, int, float]
            if len(row) != 3:
                continue
            # only use rows matching this chromosome
            if row[0] != chromosome:
                continue
            try:
                start_val = int(row[1])
                rate_val  = float(row[2])
            except ValueError:
                continue
            rec_map_data.append({'start': start_val, 'rate': rate_val})
    
    rec_map_data.sort(key=lambda x: x['start'])

    intervals = []
    if rec_map_data and rec_map_data[0]['start'] > calc_start:
        intervals.append({'start': calc_start, 'end': rec_map_data[0]['start'], 'rate': 1.0})
    
    for i, entry in enumerate(rec_map_data):
        interval_start = entry['start']
        interval_end = rec_map_data[i+1]['start'] if i < len(rec_map_data)-1 else calc_end
        if interval_end > calc_start and interval_start < calc_end:
            intervals.append({
                'start': max(interval_start, calc_start),
                'end':   min(interval_end,   calc_end),
                'rate':  entry['rate']
            })
    
    if intervals:
        last_end = intervals[-1]['end']
        if last_end < calc_end:
            intervals.append({'start': last_end, 'end': calc_end, 'rate': 1.0})
    else:
        intervals.append({'start': calc_start, 'end': calc_end, 'rate': 1.0})
    
    rec_rates = []
    num_chunks = (calc_end - calc_start + chunk_size - 1) // chunk_size
    for chunk in range(num_chunks):
        start_chunk = calc_start + chunk * chunk_size
        end_chunk   = min(calc_end, start_chunk + chunk_size)
        chunk_len   = end_chunk - start_chunk
        weighted_sum = 0.0
        
        for iv in intervals:
            o_start = max(start_chunk, iv['start'])
            o_end   = min(end_chunk,   iv['end'])
            if o_start < o_end:
                weighted_sum += iv['rate'] * (o_end - o_start)
        
        rec_rates.append(weighted_sum / chunk_len if chunk_len > 0 else 1.0)
    
    return np.array(rec_rates)



from core.calculateB import calculateB_linear

def findFlankLen(target_B, length_of_element, max_distance=1e6):
    """
    Find the minimum distance where B > target_B for a given element length,
    or return max_distance if no such distance exists.
    """
    low, high = 0, max_distance
    while low < high:
        mid = (low + high) // 2
        B = calculateB_linear(mid, length_of_element)
        if B > target_B:
            high = mid  # Narrow the search to smaller distances
        else:
            low = mid + 1  # Narrow the search to larger distances

    # After exiting the loop, check if the condition is met
    final_B = calculateB_linear(low, length_of_element)
    if final_B > target_B:
        return int(low)
    else:
        return int(max_distance)  # Return max_distance if no valid distance is found
    


## Population genetic parameters for the simulated or empirical population
## Accurate estimation requires accurate and appropriate parameters
##
## e.g. Bvalcalc --params path/to/ExampleParams.py
##
## Core parameters
f = 0.9 # Inbreeding coefficient, selfing (F = S/(2-S); Wright's inbreeding coefficient)
x = 1 # Scaling factor (N,u,r), keep as 1 unless calculating for rescaled simulations
Nanc = 80000 / (1+f) / x # Ancestral population size
r = 7.4e-7 * (1-f) * x # Recombination (crossover) rate per bp, per generation (sex-averaged)
u = 5.6e-9 * x # Mutation rate (all types) per bp, per generation
g = 3.6e-7 * (1-f) * x # Gene conversion initiation rate per bp, per generation
k = 1400 # Gene conversion tract length (bp)
# ## DFE parameters (Sum must equal 1)
f0 = 0.52 # Proportion of effectively neutral mutations with 0 <= |2Ns| < 1 (Note that 2Ns<5 does not contribute to BGS) [Naive value]
f1 = 0.16 # Proportion of weakly deleterious mutations with 1 <= |2Ns| < 10 [Naive value]
f2 = 0.16 # Proportion of moderately deleterious mutations with 10 <= |2Ns| < 100 [Naive value]
f3 = 0.16 # Proportion of strongly deleterious mutations with |2Ns| >= 100 [Naive value]
# ## Demography parameters
Ncur = Nanc # Current population size (!Requires --pop_change)
time_of_change = 0.1 # Time in Nanc generations ago that effective population size went from Nanc to Ncur (!Requires --pop_change)
## Advanced DFE parameters 
h = 0.5 + (f-0.5*f) # Dominance coefficient of selected alleles
mean, shape = 500, 0.5 # Gamma distribution of DFE to discretize and replace f0-f3 [mean (2Ns), shape] (!Requires --gamma_dfe)
## Literature cited
# [1]
# [2]
# [3]
# [4]



# def calculateB_linear_oldGC(distance_to_element, length_of_element):
#     """
#         Not currently used in main Bvalcalc function, 
#         Here for reference, Parul's gene conversion equation.
#     """    
#     C = (1.0 - np.exp(-2.0 * r * distance_to_element)) / 2.0 # cM
#     U = length_of_element * u
#     if g == 0:
#         a = C
#         b = C + (r * length_of_element) # cM
#     elif g > 0:
#         threshold = distance_to_element + length_of_element < 0.5 * k # Arbitrary threshold
#         a = np.where(
#             threshold, 
#             C + (g * distance_to_element), #pGC happens outside the element to BREAK linkage, lowers because sometimes both GC
#             C + g * k #pGC happens outside element to BREAK linkage
#             ) 
#         b = np.where(
#             threshold,
#             C + (r * length_of_element) + (g * (distance_to_element + length_of_element)), #pGC happens within element to BREAK linkage, IS IT LOWER???
#             C + (r * length_of_element) + (g * k) #pGC happens within element to BREAK linkage,
#         )

#     E_f1 = calculate_exponent(t1half, t2, U, a, b)
#     E_f2 = calculate_exponent(t2, t3, U, a, b)
#     E_f3 = calculate_exponent(t3, t4, U, a, b)

#     E_bar = ( # Sum over the DFE
#         f0 * 0.0
#         + f1 * ((t1half - t1) / (t2 - t1)) * 0.0
#         + f1 * ((t2 - t1half) / (t2 - t1)) * E_f1
#         + f2 * E_f2
#         + f3 * E_f3)

#     return np.exp(-1.0 * E_bar) # Return B



from Bvalcalc.core.helpers.demography_helpers import get_Bcur
from Bvalcalc.utils.bin_outputs import bin_outputs
import numpy as np
import os

def geneBcalc(args):    
    element_size, flank_len, quiet = args.element_size, args.flank_len, args.quiet
    
    import Bvalcalc.utils.dfe_helper as dfe_helper
    dfe_helper.CONSTANT_DFE = args.constant_dfe # Update DFE if --constant_dfe
    dfe_helper.GAMMA_DFE = args.gamma_dfe # Update DFE if --gamma_dfe
    from Bvalcalc.core.calculateB import calculateB_linear # Import calculateB with updated gamma DFE if needed

    if not quiet: 
        print(f"====== P A R A M E T E R S =========================")
        print(f"Distribution of fitness effects (DFE): {flank_len}bp")
        print(f"Length of element under selection: {element_size}bp")
        print(f"Length of flanking neutral region: {flank_len}bp")

    print(f"====== S T A R T I N G ===== C A L C ===============")
    b_values = calculateB_linear(np.arange(1, flank_len, 1, dtype = int), element_size) # B for flank region

    print(f"====== F I N I S H E D ===== C A L C ===============")

    if not quiet:
        print(f"====== R E S U L T S ! =============================")

    if args.pop_change:
        if not quiet: print("B prior to demographic change", b_values)
        b_values = get_Bcur(b_values)
        if not quiet: print("B post B-calculation", b_values)
    output_data = np.column_stack((np.arange(1, flank_len, 1, dtype = int), b_values))

    if not quiet:
        print(f"B for adjacent site: {b_values[0]}")
        print(f"Mean B for flanking region: {b_values.mean()}")
        print(f"B at start and end of the neutral region: {b_values}")

    if args.out is not None: # Write to CSV
        # 1) decouple
        positions = output_data[:, 0].astype(int)
        bvals     = output_data[:, 1].astype(float)
        b_bvals, b_pos = bin_outputs(bvals, positions, args.out_binsize) # 2) bin them
        binned_output = np.column_stack((b_pos, b_bvals)) # 3) rebuild a two-column array for the binned output
        
        np.savetxt(args.out, # This might be "b_values.csv" or a custom path
            binned_output, delimiter=",", header="Distance,B", fmt=("%d", "%.6f"), comments="")
        print(f"Saved B values to: {os.path.abspath(args.out)}")
    else:
        if not args.quiet:
            print("No output CSV requested; skipping save.")
    
    ## !!The following (prior to return) calculates output_data WITH the sites in the gene from the edge to the middle of the gene
    ## !!Not currently supported because it'll require major updates to plotting --gene, worth adding when plotB is refactored

        # distance_to_middle_of_gene = int((element_size)/2)
        # left_gene_lengths = np.arange(distance_to_middle_of_gene, distance_to_middle_of_gene*2, 1, dtype = int)
        # right_gene_lengths = np.arange(1, distance_to_middle_of_gene + 1, 1, dtype = int)[::-1] #Flip array at the end so it starts with center point

        # right_gene_b_values = calculateB_linear(distance_to_element = 0, length_of_element = right_gene_lengths) 
        # ##Distance = 0 includes BGS from focal site, doesn't make logical sense but avoids weird increase in B as selected elements have 1 less site when calculating within them
        # left_gene_b_values = calculateB_linear(distance_to_element = 1, length_of_element = left_gene_lengths)
        
        # in_gene_b_values = right_gene_b_values*left_gene_b_values
        # print(element_size, distance_to_middle_of_gene, np.shape(left_gene_b_values), np.shape(right_gene_b_values), left_gene_lengths, right_gene_lengths, in_gene_b_values)

        # combined_b_values = np.append(in_gene_b_values, b_values)
        # print(combined_b_values, np.shape(combined_b_values))
        # with_gene_output_data = np.column_stack((np.arange(1 - distance_to_middle_of_gene, len(combined_b_values) + 1 - distance_to_middle_of_gene, 1, dtype = int), combined_b_values))

        # print("Kaizo", output_data, with_gene_output_data)

    return output_data



## Population genetic parameters for the simulated or empirical population
## Accurate estimation requires accurate and appropriate parameters
##
## e.g. Bvalcalc --params path/to/ExampleParams.py
##
## Core parameters
x = 1 # Scaling factor (N,u,r), keep as 1 unless calculating for rescaled simulations
Nanc = 7300 / x # Ancestral population size [1]
r = 1e-8 * x # Recombination (crossover) rate per bp, per generation (sex-averaged) [2]
u = 1.25e-8 * x # Mutation rate (all types) per bp, per generation [2]
g = 2e-8 * x # Gene conversion initiation rate per bp, per generation [3]
k = 459 # Gene conversion tract length (bp) [3]
## DFE parameters for ALL sites in annotated regions (Sum must equal 1)
f0 = 0.51 # Proportion of effectively neutral mutations with 0 <= |2Ns| < 1 (Note that 2Ns<5 does not contribute to BGS) [4]
f1 = 0.14 # Proportion of weakly deleterious mutations with 1 <= |2Ns| < 10 [4]
f2 = 0.14 # Proportion of moderately deleterious mutations with 10 <= |2Ns| < 100 [4]
f3 = 0.21 # Proportion of strongly deleterious mutations with |2Ns| >= 100 [4]
## Demography parameters
Ncur = 14474 # Current population size (!Requires --pop_change) [1]
time_of_change = 0.81 # Time in Nanc generations ago that effective population size went from Nanc to Ncur (!Requires --pop_change) [1]
## Advanced DFE parameters 
h = 0.5 # Dominance coefficient of selected alleles [Naive value]
mean, shape = 500, 0.5 # Gamma distribution of DFE to discretize and replace f0-f3 [mean (2Ns), shape] (!Requires --gamma_dfe) [Naive value]
## Literature cited
# [1] Gutenkunst et al 2009 doi: 10.1371/journal.pgen.1000695
# [2] Kong et al 2012 doi: 10.1038/nature11396
# [3] Masaki and Browning 2025 doi: 10.1101/2024.12.30.630818
# [4] Huber et al 2017 doi: 10.1073/pnas.1619508114


from Bvalcalc.core.calculateB import calculateB_linear, calculateB_recmap
from Bvalcalc.core.helpers.calc_B_from_chunks import calc_B_from_chunks
from Bvalcalc.core.helpers.calc_R_len_dist import calc_R_lengths
from Bvalcalc.core.helpers.calc_R_len_dist import calc_R_distances
from Bvalcalc.core.helpers.calc_B_in_genes import calc_B_in_genes
from Bvalcalc.core.helpers.calc_B_precise_noninterfering import calc_B_precise_noninterfering
import numpy as np

def process_single_chunk(chunk_idx, chunk_size, blockstart, blockend, chr_start, chr_size,
                         calc_start, calc_end, num_chunks, precise_chunks,lperchunk, 
                         b_values, rec_rate_per_chunk=None, gc_rate_per_chunk=None, hri=False, quiet=False, verbose=False, unlinked_B=1.0):
    
    chunk_start =  chr_start + chunk_idx * chunk_size
    chunk_end   = min(chunk_start + chunk_size - 1, calc_end)
    chunk_slice = b_values[chunk_start:chunk_end+1] # Get b_values for this chunk
    pos_chunk = np.arange(chunk_start, chunk_end+1) # Array of positions in this chunk
    
    # Identify blocks in the "precise region"
    precise_region_start = np.maximum(chr_start, chr_start + (chunk_idx - precise_chunks) * chunk_size)
    precise_region_end   = np.minimum(chr_size, chr_start + (chunk_idx + 1 + precise_chunks) * chunk_size - 1)
    precise_blockregion_mask = (
        (precise_region_end   > blockstart) &
        (precise_region_start < blockend))
    precise_blockstart = np.clip(blockstart[precise_blockregion_mask],a_min=precise_region_start, a_max=precise_region_end)
    precise_blockend   = np.clip(blockend[precise_blockregion_mask], a_min=precise_region_start, a_max=precise_region_end)

    physical_distances_upstream   = pos_chunk[None, :] - precise_blockend[:, None] # All distances to blockends (upstream and downstream)
    physical_distances_downstream = precise_blockstart[:, None] - pos_chunk[None, :] # All distances to blockstarts (upstream and downstream)

    downstream_mask = (pos_chunk < precise_blockstart[:, None]) # True when position is less than blockstart (gene is downstream) 
    upstream_mask   = (pos_chunk > precise_blockend[:, None]) # True when position is more than blockend (gene is upstream)
    flanking_mask   = downstream_mask | upstream_mask 

    physical_distances = np.where( # Filter so only distances to blockends upstream and blockstarts downstream kept
        flanking_mask,
        np.where(upstream_mask, physical_distances_upstream, physical_distances_downstream),
        np.nan
    )
    flat_distances = physical_distances[flanking_mask] # Flatten array

    physical_lengths = precise_blockend - precise_blockstart
    flat_lengths   = np.repeat(physical_lengths, flanking_mask.sum(axis=1))
    nonzero_mask = flat_lengths != 0 # Remove genes of length 0
    flat_distances = flat_distances[nonzero_mask]
    flat_lengths   = flat_lengths[nonzero_mask]
    
    if rec_rate_per_chunk is not None: # IF REC_RATE MAP IS AVAILABLE 
        precise_rates = rec_rate_per_chunk[np.maximum(0, chunk_idx - precise_chunks):np.minimum(num_chunks, chunk_idx + precise_chunks + 1)]
        rec_lengths = calc_R_lengths(precise_blockstart , precise_blockend, precise_rates, precise_region_start, precise_region_end, chunk_size)
        rec_distances_upstream, rec_distances_downstream = calc_R_distances(precise_blockstart, precise_blockend, precise_rates, precise_region_start, precise_region_end, chunk_size, pos_chunk, chunk_start)
        rec_distances = np.where(
            flanking_mask,
            np.where(upstream_mask, rec_distances_upstream, rec_distances_downstream),
            np.nan)
        flat_rec_distances = rec_distances[flanking_mask]
        flat_rec_lengths   = np.repeat(rec_lengths, flanking_mask.sum(axis=1))
        nonzero_rec_mask = flat_lengths != 0 # Remove genes of length 0
        flat_rec_distances = flat_rec_distances[nonzero_rec_mask]
        flat_rec_lengths   = flat_rec_lengths[nonzero_rec_mask]

    if gc_rate_per_chunk is not None: # IF GC_RATE MAP IS AVAILABLE 
        precise_gc_rates = gc_rate_per_chunk[np.maximum(0, chunk_idx - precise_chunks):np.minimum(num_chunks, chunk_idx + precise_chunks + 1)]
        gc_lengths = calc_R_lengths(precise_blockstart, precise_blockend, precise_gc_rates, precise_region_start, precise_region_end, chunk_size)
        gc_distances_upstream, gc_distances_downstream = calc_R_distances(precise_blockstart, precise_blockend, precise_gc_rates, precise_region_start, precise_region_end, chunk_size, pos_chunk, chunk_start)
        gc_distances = np.where(
            flanking_mask,
            np.where(upstream_mask, gc_distances_upstream, gc_distances_downstream),
            np.nan)
        flat_gc_distances = gc_distances[flanking_mask]
        flat_gc_lengths   = np.repeat(gc_lengths, flanking_mask.sum(axis=1))
        nonzero_gc_mask = flat_lengths != 0 # Remove genes of length 0
        flat_gc_distances = flat_gc_distances[nonzero_gc_mask]
        flat_gc_lengths   = flat_gc_lengths[nonzero_gc_mask]

    B_from_distant_chunks = calc_B_from_chunks( # Compute B from distant chunks in non-precise region
        chunk_idx, chunk_size, chr_start, chr_size, num_chunks, 
        precise_chunks, lperchunk, rec_rate_per_chunk, gc_rate_per_chunk)
    
    # Calculate B for genes within chunk. within_gene_B is to include B for genic sites from BGS caused by the gene they're in
    if rec_rate_per_chunk is not None and gc_rate_per_chunk is not None: # IF REC_RATE MAP IS AVAILABLE and GC IS AVAILABLE
        within_gene_B = calc_B_in_genes(chunk_size, num_chunks, precise_chunks, precise_blockstart, precise_blockend, chunk_start, chunk_end+1, physical_lengths, precise_region_start, chunk_idx, rec_rate_per_chunk = rec_rate_per_chunk, gc_rate_per_chunk = gc_rate_per_chunk, rec_lengths = rec_lengths, gc_lengths = gc_lengths)
        flank_B = calculateB_recmap(distance_to_element=flat_distances, length_of_element=flat_lengths, rec_distances=flat_rec_distances, 
                                    rec_lengths=flat_rec_lengths, gc_distances=flat_gc_distances, gc_lengths=flat_gc_lengths)
    elif rec_rate_per_chunk is not None and gc_rate_per_chunk is None: # IF REC_RATE MAP IS AVAILABLE and GC NOT AVAILABLE
        within_gene_B = calc_B_in_genes(chunk_size, num_chunks, precise_chunks, precise_blockstart, precise_blockend, chunk_start, chunk_end+1, physical_lengths, precise_region_start, chunk_idx, rec_rate_per_chunk = rec_rate_per_chunk, gc_rate_per_chunk = None, rec_lengths = rec_lengths, gc_lengths = None)
        flank_B = calculateB_recmap(distance_to_element=flat_distances, length_of_element=flat_lengths, 
                                    rec_distances=flat_rec_distances, rec_lengths=flat_rec_lengths)
    elif rec_rate_per_chunk is None and gc_rate_per_chunk is not None: # IF REC_RATE MAP NOT AVAILABLE and GC IS AVAILALBE
        within_gene_B = calc_B_in_genes(chunk_size, num_chunks, precise_chunks, precise_blockstart, precise_blockend, chunk_start, chunk_end+1, physical_lengths, precise_region_start, chunk_idx, rec_rate_per_chunk = None, gc_rate_per_chunk = gc_rate_per_chunk, rec_lengths = None, gc_lengths = gc_lengths)
        flank_B = calculateB_recmap(distance_to_element=flat_distances, length_of_element=flat_lengths, 
                                    rec_distances=None, rec_lengths=None, gc_distances=flat_gc_distances, gc_lengths=flat_gc_lengths)
    else: # NO MAPS AVAILABLE
        within_gene_B = calc_B_in_genes(chunk_size, num_chunks, precise_chunks, precise_blockstart, precise_blockend, chunk_start, chunk_end+1, physical_lengths, precise_region_start, chunk_idx, rec_rate_per_chunk = None, gc_rate_per_chunk = None, rec_lengths = None, gc_lengths = None)
        flank_B = calculateB_linear(flat_distances, flat_lengths)

    # Combine B's calculated from distant genes, and genes within the region!
    safe_flank_B = np.concatenate((np.ones(chunk_end+1 - chunk_start, dtype=float), flank_B)) # Add an array of flank_B where all sites are B = 1, to account for sites with no flanking genes
    new_flanking_mask = np.concatenate((np.ones((1, chunk_end+1 - chunk_start), dtype=bool), flanking_mask), axis=0)
    unique_indices, inverse_indices = np.unique(np.where(new_flanking_mask)[1], return_inverse=True)
    aggregated_B = np.ones_like(np.ones_like(np.arange(chunk_start,chunk_end+1), dtype=np.float64), dtype=np.float64)
    np.multiply.at(aggregated_B, inverse_indices, safe_flank_B) # Multiplicative sum of B calculated at a given site from multiple elements
    hri_r_threshold = 0.1 # fraction of "r" in a chunk that triggers Bprime hri calculation
    hri_L_threshold = 1000 # minimum number of selected sites in a chunk that triggers Bprime hri calculation
    
    # Check if this chunk should trigger HRI calculation
    should_do_hri = False
    if (hri and rec_rate_per_chunk is not None and rec_rate_per_chunk[chunk_idx] < hri_r_threshold):
        # Calculate the total L across the entire interference region
        low_rec_chunk_ids = rec_rate_per_chunk < hri_r_threshold
        interference_region_start_idx, interference_region_end_idx = chunk_idx, chunk_idx
        while interference_region_start_idx > 0 and low_rec_chunk_ids[interference_region_start_idx - 1]:
            interference_region_start_idx -= 1
        while interference_region_end_idx < low_rec_chunk_ids.size - 1 and low_rec_chunk_ids[interference_region_end_idx + 1]:
            interference_region_end_idx += 1
        total_interfering_L = lperchunk[interference_region_start_idx : interference_region_end_idx + 1].sum()
        should_do_hri = total_interfering_L > hri_L_threshold
    
    if should_do_hri: # Only do this if user has --hri active
        from Bvalcalc.core.helpers.calc_B_in_hri_region import calc_B_in_hri_region
        hri_aggregated_B = calc_B_in_hri_region(quiet, chunk_idx, rec_rate_per_chunk, hri_r_threshold, lperchunk, chunk_size, chr_start, chr_size, num_chunks, gc_rate_per_chunk, precise_chunks, precise_blockstart, precise_blockend, pos_chunk, chunk_end, precise_region_start, precise_region_end, unlinked_B)
        chunk_slice *= hri_aggregated_B
        return b_values
    else:
        if unique_indices.size == 0: # If there are no nearby sites under selection
            chunk_slice *= (B_from_distant_chunks * within_gene_B)
        else:
            chunk_slice *= (aggregated_B * B_from_distant_chunks * within_gene_B) # Update chunk slice and combine flank_B with B from distant chunks

        mean_chunk_b = np.nanmean(chunk_slice) # Mean B for chunk




    if verbose: # Per-chunk summaries
        print(f"Processing chunk {chunk_idx}: {pos_chunk.min()} - {pos_chunk.max()}")
        if rec_rate_per_chunk is not None:
            print(f"Chunk {chunk_idx}: recombination rate = {rec_rate_per_chunk[chunk_idx]}")
        if gc_rate_per_chunk is not None:
            print(f"Chunk {chunk_idx}: recombination rate = {gc_rate_per_chunk[chunk_idx]}")
        # print(f"B from distant chunks: {B_from_distant_chunks}")
        print(f"Number of relevant genes: {len(precise_blockstart)}")
        print(f"Number of neutral sites in chunk [{chunk_start}-{chunk_end}): {np.isnan(chunk_slice).sum()}")
        # print(f"Aggregated B values for chunk: {aggregated_B}")
        print(f"Mean B value for chunk {chunk_idx}: [{chunk_start}-{chunk_end}]: {mean_chunk_b}")

    return b_values


import numpy as np
from Bvalcalc.core.calculateB import calculateB_linear, calculateB_recmap

def calc_B_in_genes(
    chunk_size, num_chunks, precise_chunks,
    precise_blockstart, precise_blockend,
    chunk_start, chunk_end,
    physical_lengths, precise_region_start, chunk_num,
    rec_rate_per_chunk=None, gc_rate_per_chunk=None,
    rec_lengths=None, gc_lengths=None
):
    genes_in_this_chunk_mask = np.logical_and(
        precise_blockstart < chunk_end,
        precise_blockend >= chunk_start
    )
    this_chunk_blockstart = precise_blockstart[genes_in_this_chunk_mask]
    this_chunk_blockend = precise_blockend[genes_in_this_chunk_mask]

    if len(this_chunk_blockstart) == 0:
        return np.ones(chunk_end - chunk_start, dtype=np.float64)

    this_chunk_blockstart_inchunk = np.clip(this_chunk_blockstart, chunk_start, chunk_end - 1)
    this_chunk_blockend_inchunk = np.clip(this_chunk_blockend, chunk_start, chunk_end - 1)

    within_gene_B = np.ones(chunk_end - chunk_start, dtype=np.float64) # Initialize array
    chunk_starts = precise_region_start + np.arange(num_chunks + 1) * chunk_size
    this_chunk_idx = np.searchsorted(chunk_starts, chunk_start)

    # Pre-slice once
    masked_phys_lengths = physical_lengths[genes_in_this_chunk_mask]
    masked_rec_lengths = rec_lengths[genes_in_this_chunk_mask] if rec_lengths is not None else None
    masked_gc_lengths = gc_lengths[genes_in_this_chunk_mask] if gc_lengths is not None else None

    # Precompute rate slices
    precise_slice_start = max(0, chunk_num - precise_chunks)
    precise_slice_end = min(num_chunks, chunk_num + precise_chunks + 1)

    if rec_rate_per_chunk is not None:
        precise_rates = rec_rate_per_chunk[precise_slice_start:precise_slice_end]
        rec_bp_to_element = precise_rates[this_chunk_idx]

    if gc_rate_per_chunk is not None:
        precise_gc_rates = gc_rate_per_chunk[precise_slice_start:precise_slice_end]
        gc_bp_to_element = precise_gc_rates[this_chunk_idx]

    for gene_idx in range(len(this_chunk_blockstart_inchunk)): # Ragged array needs for loop
        start = this_chunk_blockstart_inchunk[gene_idx]
        end = this_chunk_blockend_inchunk[gene_idx] + 1
        gpos_in_chunk = np.arange(start, end)
        gene_sites = gpos_in_chunk - chunk_start

        left_block_lengths = gpos_in_chunk - this_chunk_blockstart[gene_idx]
        right_block_lengths = this_chunk_blockend[gene_idx] - gpos_in_chunk
        focal_block_physlength = masked_phys_lengths[gene_idx]

        if rec_rate_per_chunk is not None and gc_rate_per_chunk is None:
            focal_block_reclength = masked_rec_lengths[gene_idx]
            scale = focal_block_reclength / focal_block_physlength
            left_chunk_reclengths = left_block_lengths * scale
            right_chunk_reclengths = right_block_lengths * scale

            left_block_B = calculateB_recmap(1, left_block_lengths, rec_bp_to_element, left_chunk_reclengths)
            right_block_B = calculateB_recmap(1, right_block_lengths, rec_bp_to_element, right_chunk_reclengths)

        elif rec_rate_per_chunk is not None and gc_rate_per_chunk is not None:
            focal_block_reclength = masked_rec_lengths[gene_idx]
            focal_block_gclength = masked_gc_lengths[gene_idx]
            # Fix needed here for RuntimeWarning: invalid value encountered in scalar divide
            rec_scale = focal_block_reclength / focal_block_physlength
            gc_scale = focal_block_gclength / focal_block_physlength

            left_chunk_reclengths = left_block_lengths * rec_scale
            right_chunk_reclengths = right_block_lengths * rec_scale

            left_chunk_gclengths = left_block_lengths * gc_scale
            right_chunk_gclengths = right_block_lengths * gc_scale

            left_block_B = calculateB_recmap(
                1, left_block_lengths, rec_bp_to_element, left_chunk_reclengths,
                gc_bp_to_element, left_chunk_gclengths
            )
            right_block_B = calculateB_recmap(
                1, right_block_lengths, rec_bp_to_element, right_chunk_reclengths,
                gc_bp_to_element, right_chunk_gclengths
            )

        elif rec_rate_per_chunk is None and gc_rate_per_chunk is not None:
            focal_block_gclength = masked_gc_lengths[gene_idx]
            gc_scale = focal_block_gclength / focal_block_physlength

            left_chunk_gclengths = left_block_lengths * gc_scale
            right_chunk_gclengths = right_block_lengths * gc_scale

            left_block_B = calculateB_recmap(1, left_block_lengths, None, None, gc_bp_to_element, left_chunk_gclengths)
            right_block_B = calculateB_recmap(1, right_block_lengths, None, None, gc_bp_to_element, right_chunk_gclengths)

        else:
            left_block_B = calculateB_linear(1, left_block_lengths)
            right_block_B = calculateB_linear(1, right_block_lengths)

        np.multiply.at(within_gene_B, gene_sites, left_block_B)
        np.multiply.at(within_gene_B, gene_sites, right_block_B)

    return within_gene_B



"""
Helper modules for core calculations in bvalcalc.
"""
from .calc_B_from_chunks import calc_B_from_chunks
from .calc_B_in_genes   import calc_B_in_genes
from .calc_L_per_chunk  import calculate_L_per_chunk
from .calc_R_len_dist   import calc_R_lengths, calc_R_distances, calc_R_lendist_for_chunks
from .demography_helpers import get_Bcur
from .process_single_chunk import process_single_chunk

__all__ = [
    "calc_B_from_chunks",
    "calc_B_in_genes",
    "calculate_L_per_chunk",
    "calc_R_lengths",
    "calc_R_distances",
    "get_Bcur",
    "process_single_chunk",
]



from Bvalcalc.core.calculateB import calculateB_linear, calculateB_recmap
from Bvalcalc.core.helpers.calc_R_len_dist import calc_R_lengths, calc_R_distances
import numpy as np

def calc_B_precise_noninterfering(
    precise_blockstart, precise_blockend, pos_chunk,
    chr_start, chunk_end, chunk_size, chr_size, precise_region_start, precise_region_end,
    local_interference_indices, chunk_idx,
    rec_rate_per_chunk, gc_rate_per_chunk
):
    """
    Calculate B in the precise region excluding local interference blocks,
    with optional recombination/gene-conversion formulas.

    DEV note: Lots of code is redundant here with the parent process_single_chunk().
    DEV note: Could be streamlined but given HRI calculation is assumed to be relatively
    rare, the speed hit should be manageable.
    """
    # 1) bp range of interference region
    start_bp = chr_start + local_interference_indices[0] * chunk_size
    end_bp   = min(
        chr_start + (local_interference_indices[-1] + 1) * chunk_size - 1,
        chr_size
    )
    # print(f"[DEBUG] interference bp range: {start_bp}-{end_bp}")

    # 2) trim blocks to remove overlap
    ts, te = [], []
    for bs, be in zip(precise_blockstart, precise_blockend):
        if be < start_bp or bs > end_bp:
            ts.append(bs); te.append(be)
        else:
            if bs < start_bp:
                ts.append(bs); te.append(start_bp - 1)
            if be > end_bp:
                ts.append(end_bp + 1); te.append(be)
    # print(f"[DEBUG] original blocks: {list(zip(precise_blockstart, precise_blockend))}")
    # if not ts:
    #     print(f"Chunk {chunk_idx}: No non-interfering blocks; B=1")
    #     return 1.0
    bs_arr = np.array(ts, dtype=int)
    be_arr = np.array(te, dtype=int)
    # print(f"[DEBUG] trimmed blocks:  {list(zip(bs_arr, be_arr))}")

    # 3) build distance mask
    up_dist   = pos_chunk[None, :] - be_arr[:, None]
    down_dist = bs_arr[:, None] - pos_chunk[None, :]
    up_mask   = pos_chunk > be_arr[:, None]
    down_mask = pos_chunk < bs_arr[:, None]
    mask      = up_mask | down_mask

    phys = np.where(mask, np.where(up_mask, up_dist, down_dist), np.nan)
    flat_distances = phys[mask]

    # print(f"[DEBUG] flat_distances ({len(flat_distances)}): "
    #       f"{flat_distances[:5]}{'...' if len(flat_distances)>5 else ''}")

    # 4) flatten lengths
    lengths = be_arr - bs_arr
    counts  = mask.sum(axis=1)
    flat_lengths = np.repeat(lengths, counts)
    # print(f"[DEBUG] flat_lengths   ({len(flat_lengths)}): "
    #       f"{flat_lengths[:5]}{'...' if len(flat_lengths)>5 else ''}")

    # 5) drop zeros
    valid = flat_lengths > 0
    flat_distances = flat_distances[valid]
    flat_lengths   = flat_lengths[valid]

    # 6) optionally flatten recombination
    if rec_rate_per_chunk is not None:
        region_start_idx = (precise_region_start.min() - chr_start) // chunk_size
        region_end_idx   = (precise_region_end.max() - chr_start) // chunk_size
        r_rates = rec_rate_per_chunk[region_start_idx:region_end_idx+1]

        rec_lens = calc_R_lengths(
            bs_arr, be_arr,
            r_rates,
            precise_region_start, precise_region_end,
            chunk_size
        )
        chunk_start = chr_start + chunk_idx * chunk_size
        rec_up, rec_down = calc_R_distances(
            bs_arr, be_arr,
            r_rates,
            precise_region_start, precise_region_end,
            chunk_size,
            pos_chunk,
            chunk_start
        )
        rec_phys = np.where(mask, np.where(up_mask, rec_up, rec_down), np.nan)
        flat_rec_distances = rec_phys[mask][valid]
        flat_rec_lengths   = np.repeat(rec_lens, counts)[valid]
        # print(f"[DEBUG] flat_rec_distances ({len(flat_rec_distances)}): "
        #       f"{flat_rec_distances[:5]}{'...' if len(flat_rec_distances)>5 else ''}")
        # print(f"[DEBUG] flat_rec_lengths    ({len(flat_rec_lengths)}):  "
        #       f"{flat_rec_lengths[:5]}{'...' if len(flat_rec_lengths)>5 else ''}")
    else:
        flat_rec_distances = flat_rec_lengths = None

    # 7) optionally flatten gene-conversion
    if gc_rate_per_chunk is not None:
        region_start_idx = (precise_region_start - chr_start) // chunk_size
        region_end_idx   = (precise_region_end - chr_start) // chunk_size
        g_rates = gc_rate_per_chunk[region_start_idx:region_end_idx+1]

        gc_lens = calc_R_lengths(
            bs_arr, be_arr,
            g_rates,
            precise_region_start, precise_region_end,
            chunk_size
        )
        chunk_start = chr_start + chunk_idx * chunk_size
        gc_up, gc_down = calc_R_distances(
            bs_arr, be_arr,
            g_rates,
            precise_region_start, precise_region_end,
            chunk_size,
            pos_chunk,
            chunk_start
        )
        gc_phys = np.where(mask, np.where(up_mask, gc_up, gc_down), np.nan)
        flat_gc_distances = gc_phys[mask][valid]
        flat_gc_lengths   = np.repeat(gc_lens, counts)[valid]
       # print(f"[DEBUG] flat_gc_distances ({len(flat_gc_distances)}): "
         #     f"{flat_gc_distances[:5]}{'...' if len(flat_gc_distances)>5 else ''}")
        #print(f"[DEBUG] flat_gc_lengths    ({len(flat_gc_lengths)}):  "
          #    f"{flat_gc_lengths[:5]}{'...' if len(flat_gc_lengths)>5 else ''}")
    else:
        flat_gc_distances = flat_gc_lengths = None

    # 8) dispatch to the correct B function
    if flat_rec_distances is not None and flat_gc_distances is not None:
        B = calculateB_recmap(
            distance_to_element=flat_distances,
            length_of_element=flat_lengths,
            rec_distances=flat_rec_distances,
            rec_lengths=flat_rec_lengths,
            gc_distances=flat_gc_distances,
            gc_lengths=flat_gc_lengths
        )
    elif flat_rec_distances is not None:
        B = calculateB_recmap(
            distance_to_element=flat_distances,
            length_of_element=flat_lengths,
            rec_distances=flat_rec_distances,
            rec_lengths=flat_rec_lengths
        )
    elif flat_gc_distances is not None:
        B = calculateB_recmap(
            distance_to_element=flat_distances,
            length_of_element=flat_lengths,
            gc_distances=flat_gc_distances,
            gc_lengths=flat_gc_lengths
        )
    else:
        B = calculateB_linear(flat_distances, flat_lengths)


        # Combine B's calculated from distant genes, and genes within the region!
    safe_flank_B = np.concatenate((np.ones(chunk_end+1 - chunk_start, dtype=float), B)) # Add an array of flank_B where all sites are B = 1, to account for sites with no flanking genes
    new_flanking_mask = np.concatenate((np.ones((1, chunk_end+1 - chunk_start), dtype=bool), mask), axis=0)
    unique_indices, inverse_indices = np.unique(np.where(new_flanking_mask)[1], return_inverse=True)
    aggregated_B = np.ones_like(np.ones_like(np.arange(chunk_start,chunk_end+1), dtype=np.float64), dtype=np.float64)
    np.multiply.at(aggregated_B, inverse_indices, safe_flank_B) # Multiplicative sum of B calculated at a given site from multiple elements

    print(f"Chunk {chunk_idx}: Final B_noninterfering = {aggregated_B}")
    return aggregated_B

## WON'T WORK FOR UNLINKED B, OR FOR SITES considering B from an interfering region, that aren't themselves in the interfering region



"""
Utility modules for bvalcalc.
"""
# Argument parsers
from .parseArgs import parse_args, parseGenomeArgs, parseRegionArgs, parseGeneArgs, parseSiteArgs
# Parameter generation
from .generateParams import SPECIES, generateParams
# File and map handlers
from .load_bed_gff     import load_bed_gff
from .bin_outputs import bin_outputs
from .load_chr_sizes   import load_chr_sizes
from .load_rec_map     import load_rec_map
from .load_Bmap       import load_Bmap
# Sample data utilities
from .sample_data import download_sample_data, get_sample_data_dir
# DFE utilities
from . import dfe_helper

__all__ = [
    # parsers
   "parse_args", "parseGenomeArgs", "parseRegionArgs", "parseGeneArgs", "parseSiteArgs",
    # params
    "SPECIES", "generateParams",
    # handlers
    "load_bed_gff", "bin_outputs", "load_chr_sizes", "load_rec_map", "load_Bmap",
    # sample data
    "download_sample_data", "get_sample_data_dir",
    # DFE utilities
    "dfe_helper",
]


import numpy as np

def calculate_L_per_chunk(
    chunk_size,
    blockstart,
    blockend,
    chr_start,
    chr_size
):
    """
    Loop-based calculation of how much each set of blocks (flank blocks + main blocks)
    overlaps with each chunk in [chr_start, chr_size).

    """

    # Number of chunks, ensuring we cover the entire [calc_start, calc_end)
    num_chunks = (chr_size - chr_start + chunk_size - 1) // chunk_size  # ceiling division
    
    # Precompute the start and end of each chunk
    chunk_indices = np.arange(num_chunks)
    chunk_starts = chr_start + chunk_indices * chunk_size
    chunk_ends   = chr_start + (chunk_indices + 1) * chunk_size

    # Allocate array to accumulate overlap length per chunk
    chunk_sums = np.zeros(num_chunks, dtype=float)

    def accumulate_overlaps(starts, ends):
        """
        Accumulate overlaps for a set of blocks with the pre-defined chunks.
        """
        for s, e in zip(starts, ends):
            # Ignore blocks with no positive length or completely out of range
            if e <= s:
                continue
            if e <= chr_start or s >= chr_size:
                continue

            # Clamp block coordinates to [calc_start, calc_end)
            s_clamped = max(s, chr_start)
            e_clamped = min(e, chr_size)

            if e_clamped <= s_clamped:
                continue  # No valid overlap

            # Determine which chunks this block might span
            # (integer division for chunk indices)
            start_chunk_idx = (s_clamped - chr_start) // chunk_size
            end_chunk_idx   = (e_clamped - 1 - chr_start) // chunk_size

            # Ensure indices are in [0, num_chunks-1]
            start_chunk_idx = max(0, start_chunk_idx)
            end_chunk_idx   = min(num_chunks - 1, end_chunk_idx)

            # Walk through the relevant chunks and accumulate overlap
            for c in range(start_chunk_idx, end_chunk_idx + 1):
                overlap_start = max(s_clamped, chunk_starts[c])
                overlap_end   = min(e_clamped, chunk_ends[c])
                overlap_len   = overlap_end - overlap_start
                if overlap_len > 0:
                    chunk_sums[c] += overlap_len

    # 2) Accumulate overlap within given chunk for genes
    accumulate_overlaps(blockstart, blockend)

    return chunk_sums


from Bvalcalc.core.helpers.process_single_chunk import process_single_chunk
from Bvalcalc.core.helpers.calc_L_per_chunk import calculate_L_per_chunk
from Bvalcalc.core.helpers.demography_helpers import get_Bcur
from Bvalcalc.utils.load_rec_map import load_rec_map
from Bvalcalc.utils.bin_outputs import bin_outputs
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
import numpy as np
import os
import sys

def chromBcalc(args, blockstart, blockend, chromosome, unlinked_B, prior_pos = None, prior_b = None, calc_start=None, calc_end=None, chr_size=None, caller="regionBcalc"):    
    #Shared arguments between genomeBcalc and regionBcalc
    file_path, chunk_size, precise_chunks, hri, quiet, verbose = args.bedgff, args.chunk_size, args.precise_chunks, args.hri, args.quiet, args.verbose
    
    num_blocks = len(blockstart)
    # Auto-adjust chunk size for large datasets (only if user hasn't manually set chunk_size)
    if args.chunk_size is None:  # If they did not explicitly provide --chunk_size
        # Set default chunk size
        chunk_size = 20000
        original_chunk_size = chunk_size
        if num_blocks > 250000:
            chunk_size = 1000  # Use 1kb chunks for extremely massive datasets
            if not quiet:
                print(f"Extremely massive dataset detected ({num_blocks} blocks). Auto-adjusting chunk size from {original_chunk_size} to {chunk_size} bp for memory efficiency. Use --chunk_size to override.")
        elif num_blocks > 125000:
            chunk_size = 2000  # Use 2kb chunks for massive datasets
            if not quiet:
                print(f"Massive dataset detected ({num_blocks} blocks). Auto-adjusting chunk size from {original_chunk_size} to {chunk_size} bp for memory efficiency. Use --chunk_size to override.")
        elif num_blocks > 50000:
            chunk_size = 5000  # Use 5kb chunks for very large datasets
            if not quiet:
                print(f"Very large dataset detected ({num_blocks} blocks). Auto-adjusting chunk size from {original_chunk_size} to {chunk_size} bp for memory efficiency. Use --chunk_size to override.")
        elif num_blocks > 25000:
            chunk_size = 10000  # Use 10kb chunks for large datasets
            if not quiet:
                print(f"Large dataset detected ({num_blocks} blocks). Auto-adjusting chunk size from {original_chunk_size} to {chunk_size} bp for memory efficiency. Use --chunk_size to override.")
    elif not quiet and num_blocks > 25000:
        print(f"Large dataset detected ({num_blocks} blocks) but using user-specified chunk size of {chunk_size} bp.")
    #Arguments specific to regionBcalc
    if caller == "regionBcalc":
        calc_start, calc_end = calc_start, calc_end
        if calc_end > blockend[-1]:
            chr_size = calc_end
        else:
            chr_size = None

    if not args.quiet: 
        print(f"====== P A R A M E T E R S =========================")
        print(f"BED/GFF file for regions under selection: {file_path}")
        if chr_size is not None: print(f"Last position in chromosome {chromosome}: {calc_end}")
        print(f"Size of chunks to calculate B in per iteration: {chunk_size}bp")
        print(f"Number of adjacent chunks to calculate B precisely for: {precise_chunks}")

    if chr_size is not None and chr_size < blockend[-1]:
        raise ValueError(f"chr_size provided is less than gene position for chromosome {chromosome}")
    if chr_size is None: # Default chr_size to last value in blockend if not given
        if len(blockend) == 0 and caller != "regionBcalc":
            raise ValueError("chr_size was not provided for chromosome: {chromosome} and gene position ends not computed. Check BED/GFF input, and specify chr_size if needed")
        chr_size = blockend[-1]
        if calc_end is None and not args.quiet:
            print(f"No --chr_size provided for chromosome: {chromosome}. Using last position in BED/GFF: {chr_size}")

    if not quiet: print(f"====== S T A R T I N G ===== C A L C ===============")
    if calc_start is None and calc_end is None:
        if not quiet: print(f"Calculating B for entire chromosome, to only calculate for a subregion, use --calc_start and --calc_end")
    if calc_start is None:
        calc_start = 1
    if calc_end is None:
        calc_end = chr_size

    chr_start = 1 # Currently hardcoded, can change if needed
    num_chunks = (chr_size - chr_start + chunk_size - 1) // chunk_size

    calc_chunk_start = (calc_start - chr_start) // chunk_size
    calc_chunk_end = (calc_end - chr_start) // chunk_size
    calc_chunks = np.arange(calc_chunk_start,calc_chunk_end + 1) # Relevant chunks to calculate B for based on calc_start and calc_end

    b_values = np.ones(chr_size + 2 - chr_start, dtype=np.float64) # Initialize array of B values

    lperchunk = calculate_L_per_chunk(chunk_size, blockstart, blockend, chr_start, chr_size) # Cumulative conserved length in each chunk

    if args.rec_map: # Process recombination map if provided
        if not quiet: print(f"Using recombination (crossover) map from {args.rec_map}")
        rec_rate_per_chunk = load_rec_map(args.rec_map, chr_start, chr_size, chunk_size, chromosome)
    else:
        rec_rate_per_chunk = None

    if args.gc_map:
        if not quiet: print(f"Using gene conversion map from {args.gc_map}")
        gc_rate_per_chunk = load_rec_map(args.gc_map, chr_start, chr_size, chunk_size, chromosome)
    else:
        gc_rate_per_chunk = None

    if verbose: print(f"====== R E S U L T S == P E R == C H U N K =========")
    elif not quiet: print(f"To print per-chunk summaries, add --verbose.")

    import gc
    BATCH_SIZE = args.chunk_batch_size
    total_chunks = len(calc_chunks)
    completed = 0

    for batch_start in range(0, total_chunks, BATCH_SIZE):
        batch = calc_chunks[batch_start : batch_start + BATCH_SIZE]
        with ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(process_single_chunk, chunk_idx,
                                chunk_size, blockstart, blockend, chr_start, chr_size, calc_start,
                                calc_end, num_chunks, precise_chunks, lperchunk, b_values,
                                rec_rate_per_chunk, gc_rate_per_chunk, hri, quiet, verbose, unlinked_B): chunk_idx
                for chunk_idx in batch
            }
            if not quiet and not verbose:
                for future in as_completed(futures):
                    completed += 1
                    progress = int((completed / total_chunks) * 100)
                    sys.stdout.write(f"\rProgress ({chromosome}): {progress}% ({completed}/{total_chunks} chunks [{chunk_size}])")
                    sys.stdout.flush()
                # After batch is done, cleanup
                print()  # Move to the next line after progress printing
                del futures
                gc.collect()

    b_values = b_values[calc_start:(calc_end+1)] # Trim b_values array to only calculated region
    b_values = b_values * unlinked_B
    # Apply prior B values as multipliers after chunk processing
    if prior_pos is not None and prior_b is not None: # Apply prior B values to ranges
        sorted_indices = np.argsort(prior_pos) # Sort positions and B values
        sorted_pos, sorted_b = prior_pos[sorted_indices], prior_b[sorted_indices]
        for i in range(len(sorted_pos)): # Apply B values from each start position to next
            start_pos, b_val = sorted_pos[i], sorted_b[i]
            end_pos = sorted_pos[i + 1] if i < len(sorted_pos) - 1 else calc_end + 1 # Next position or end
            range_start, range_end = max(start_pos, calc_start), min(end_pos, calc_end + 1) # Clip to calc region
            if range_start < range_end:
                idx_start, idx_end = range_start - calc_start, range_end - calc_start # Convert to array indices
                b_values[idx_start:idx_end] *= b_val # Multiply by prior B values
    # print('Hriii', np.shape(b_values))

    if hri and rec_rate_per_chunk is not None: # If --hri is active
        from Bvalcalc.core.helpers.extend_hri_regions_correction import extend_hri_regions_correction
        hri_extended_starts, hri_extended_ends = extend_hri_regions_correction(b_values, rec_rate_per_chunk, chunk_size, chr_start, calc_start, calc_end, hri_r_threshold = 0.1) # Extend HRI regions until B > B' to avoid sharp decrease in B at the border between normal and HRI regions. See manuscript.
    else:
        hri_extended_starts, hri_extended_ends = np.array([], dtype=int), np.array([], dtype=int)
    
    if not quiet: 
        print(f"====== F I N I S H E D ===== C A L C ===============")
        print(f"====== R E S U L T S ====== S U M M A R Y ==========")
                # Total genic bases within calc_start to calc_end
        calc_selected_length = 0
        for start, end in zip(blockstart, blockend):
            # Find overlap between this block and the calculated region
            overlap_start = max(start, calc_start)
            overlap_end = min(end, calc_end)
            if overlap_start <= overlap_end:
                calc_selected_length += (overlap_end - overlap_start + 1)
        print(f"Cumulative length of calculated region under selection: {calc_selected_length}bp "f"({round((calc_selected_length / (calc_end - calc_start + 1)) * 100, 2)}%)")
        if prior_pos is not None and prior_b is not None: # Report mean prior B if provided
            mean_prior_b = np.mean(sorted_b)
            print(f"Mean prior B value: {mean_prior_b:.6f}")
        print(f"Cumulative length of chromosome under selection: {int(sum(lperchunk))}bp ({round((sum(lperchunk)/(chr_size - chr_start + 1))*100,2)}%)")
        print(f"B from unlinked sites calculated for chromosome {chromosome}: {unlinked_B}")
        if caller == "genomeBcalc": print(f"Mean B of neutral sites across chromosome {chromosome}: {b_values[~np.isnan(b_values)].mean()}")
        elif caller == "regionBcalc": print(f"Mean B of neutral sites across specified region: {b_values[~np.isnan(b_values)].mean()}")
        if args.rec_map: # Process recombination map if provided
            print(f"Calculated using recombination (crossover) map, with rates averaged within {chunk_size}bp chunks")
        if args.gc_map: # Process recombination map if provided
            print(f"Calculated using gene conversion map, with rates averaged within {chunk_size}bp chunks")    

    block_ranges = np.column_stack((np.repeat(chromosome, blockstart.shape[0]), blockstart, blockend))

    positions = np.arange(calc_start, calc_end + 1)
    conserved = np.full_like(positions, "N", dtype="<U1")
    for start, end in zip(blockstart, blockend): # Mark conserved regions
        conserved[max(start, calc_start) - calc_start : min(end, calc_end) - calc_start + 1] = "C"

    if args.pop_change:
        b_values = get_Bcur(b_values)
        if not quiet: print("Demographic change applied to B-calculation")

    binned_b_values, binned_positions = bin_outputs(b_values, positions, args.out_binsize)
    chrom_col = np.full(binned_positions.shape, chromosome, dtype="<U20")

    output_data = np.core.records.fromarrays(
        [chrom_col,binned_positions.astype(int),binned_b_values.astype(float)],
        names='Chromosome,Start,B',formats='U20,i8,f8')

    if args.out is not None: # Write to CSVs
        print(f"Writing B output to file...")
        from Bvalcalc.utils.write_chrom_B_to_file import write_chrom_B_to_file
        write_chrom_B_to_file(args.out, output_data, quiet, hri_extended_starts, hri_extended_ends, args.out_binsize, calc_end)
        print(f"Appended B values to: {os.path.abspath(args.out)}")
    else:
        if not args.quiet:
            print("No output CSV requested; skipping save.")

    if caller == "regionBcalc":
        if rec_rate_per_chunk is not None:
            rec_rate_per_chunk_in_region = rec_rate_per_chunk[calc_start // chunk_size:] # Slice rec_rate_per_chunk from region start onward
        else: rec_rate_per_chunk_in_region = None
        return output_data, block_ranges, rec_rate_per_chunk_in_region, chunk_size
    else: #caller is genomeBcalc
        return


import csv
import os
import numpy as np
import re

def load_bed_gff(file_path):
    """
    Read a BED, CSV, GFF, GFF3, or GTF file and return arrays of block starts, block ends, and chromosomes.

    - BED: tab-delimited, cols: chrom, start, end
    - CSV: comma-delimited, cols: chrom, start, end
    - GFF/GFF3/GTF: tab-delimited, cols: chrom, source, feature, start, end, ...

    Any lines beginning with '#' or with fewer than the required columns are skipped.
    For duplicate start coordinates, only the longest block is kept.
    """
    # Determine extension and parsing rules
    _, ext = os.path.splitext(file_path)
    ext = ext.lower()

    if ext == '.bed':
        delim = '\t'
        chrom_idx, start_idx, end_idx = 0, 1, 2
        min_cols = 3
    elif ext in ('.gff', '.gff3', '.gtf'):
        delim = '\t'
        chrom_idx, start_idx, end_idx = 0, 3, 4
        min_cols = 9
    else:
        # default to CSV
        delim = ','
        chrom_idx, start_idx, end_idx = 0, 1, 2
        min_cols = 3

    # mapping from (chrom, start) -> (end, length)
    longest_blocks = {}
    skip_count = 0

    with open(file_path, 'r') as file:
        reader = csv.reader(file, delimiter=delim)
        for row in reader:
            # skip comments
            if row and row[0].startswith('#'):
                continue
            # skip too-short rows
            if len(row) < min_cols:
                skip_count += 1
                continue

            chrom = row[chrom_idx]
            try:
                start = int(row[start_idx])
                end = int(row[end_idx])
            except ValueError:
                continue

            if end < start:
                start, end = end, start
            if start == end:
                continue

            key = (chrom, start)
            length = end - start

            if key not in longest_blocks or length > (longest_blocks[key][0] - start):
                longest_blocks[key] = (end, chrom)

    if skip_count:
        raise ValueError(f"ERROR: {skip_count} malformed lines (fewer than {min_cols} columns) in {file_path}. Consider re-formatting to .BED")

    # Unpack results
    # Convert dict to list of (chrom, start, end)
    raw_blocks = [(chrom, start, end)
                  for (chrom, start), (end, _) in longest_blocks.items()]

    # Sort by chrom, start, then descending end
    raw_blocks.sort(key=lambda x: (x[0], x[1], -x[2]))

    # Remove overlapping entries (keep longest non-overlapping)
    filtered = []
    current_end = {}

    for chrom, start, end in raw_blocks:
        if chrom in current_end and start < current_end[chrom]:
            continue
        filtered.append((chrom, start, end))
        current_end[chrom] = end

    # Unpack and return
    if filtered:
        chromosomes, blockstart, blockend = zip(*filtered)
        return np.array(blockstart), np.array(blockend), np.array(chromosomes)
    else:
        return np.array([]), np.array([]), np.array([])



import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.ticker as ticker
import numpy as np
from matplotlib.collections import LineCollection
from matplotlib import gridspec  # for rec rate strip

def plotB_figures(b_values_input, caller, output_path, quiet, gene_ranges=None, neutral_only=False, rec_rates=None):
    if not quiet:
        print('====== P L O T T I N G . . . =======================ss')

    B_uncorrected = None
    B_observed = None
    legend_name_blue = "Calculated B"
    legend_name_orange = "Uncorrected B"
    legend_name_dot = "Observed B"

    # nogcBasicParams AKA Normal
    # poetry run Bvalcalc --gene --params ./tests/testparams/nogcBasicParams.py --plot /Users/jmarsh96/Desktop/Bcalc/Figures/nogcBasicParams.png
    B_observed = "/Users/jmarsh96/Desktop/Bcalc/Figures/data/all.pi"
    legend_name_blue = "Calculated"
    legend_name_dot = "Observed (simulations)"
    title_name = 'B recovery from single element with DFE'

    # gcBasicParams AKA with GC
    # poetry run Bvalcalc --gene --params ./tests/testparams/gcBasicParams.py --plot /Users/jmarsh96/Desktop/Bcalc/Figures/gcBasicParams.png
    # B_uncorrected = "/Users/jmarsh96/Desktop/Bcalc/Figures/data/nogcBasicParams.B"
    # B_observed = "/Users/jmarsh96/Desktop/Bcalc/Figures/data/40kb_gc_all.pi"
    # legend_name_blue = "Calculated (with GC)"
    # legend_name_orange = "Calculated (no GC)"
    # legend_name_dot = "Observed (simulations)"
    # title_name = 'B recovery from single element with gene conversion'

    # expand_5N_1T AKA Demography
    # poetry run Bvalcalc --gene --params tests/testparams/ExpandParams_5N_1T.py --pop_change --plot /Users/jmarsh96/Desktop/Bcalc/Figures/expand_5N_1T.png
    # B_uncorrected = "/Users/jmarsh96/Desktop/Bcalc/Figures/data/c40kb_expand_5N_1T.bvals"
    # B_observed = "/Users/jmarsh96/Desktop/Bcalc/Figures/data/OBS_Expand_5N_1T.csv"
    # legend_name_blue = "Calculated (with demography)"
    # legend_name_orange = "Calculated (no demography)"
    # legend_name_dot = "Observed (simulations)"
    # title_name = 'B recovery from single element (5X Expansion 1N_anc generations ago)'

    # SelfParams_0.9S_0.5h AKA Selfing
    # poetry run Bvalcalc --gene --params tests/testparams/SelfParams_0.9S_0.5h.py --plot /Users/jmarsh96/Desktop/Bcalc/Figures/SelfParams_0.9S_0.5h.png
    # B_uncorrected = "/Users/jmarsh96/Desktop/Bcalc/Figures/data/nogcBasicParams.B"
    # B_observed = "/Users/jmarsh96/Desktop/Bcalc/Figures/data/40kb_f0.9_h0.5_all.pi"
    # legend_name_blue = "Calculated (with selfing)"
    # legend_name_orange = "Calculated (no selfing)"
    # legend_name_dot = "Observed (simulations)"
    # title_name = 'B recovery from single element with selfing (S = 0.9)'

    # Rename to reflect the actual parameter change

    if B_uncorrected is not None: load_B_uncorrected(B_uncorrected)
    if B_observed is not None: load_B_observed(B_observed)

    # Configure fonts and styles
    mpl.rcParams['font.family'] = ['Helvetica', 'DejaVu Sans', 'Arial']
    if 'seaborn-v0_8-whitegrid' in plt.style.available:
        plt.style.use('seaborn-v0_8-whitegrid')
    else:
        print("'seaborn-whitegrid' style not found, using 'ggplot' as fallback.")
        plt.style.use('ggplot')
        mpl.rcParams['axes.facecolor'] = 'white'
        mpl.rcParams['figure.facecolor'] = 'white'
        mpl.rcParams['grid.color'] = 'grey'
        mpl.rcParams['grid.linestyle'] = '--'
        mpl.rcParams['grid.linewidth'] = 0.5
    mpl.rcParams['axes.edgecolor'] = 'black'

    # Create figure and axis based on rec_rates
    if rec_rates is None:
        fig, ax = plt.subplots(figsize=(10, 6))
    else:
        height_ratios = [10, 0.5]
        fig = plt.figure(figsize=(10, 6))
        gs = gridspec.GridSpec(nrows=2, ncols=1, height_ratios=height_ratios)
        ax = fig.add_subplot(gs[0])

    # Main plotting logic
    if caller == "chromosome":
        positions = b_values_input['Position']
        b_vals = b_values_input['B']
        chrom = b_values_input['Chromosome'][0] if 'Chromosome' in b_values_input.dtype.names else 'unknown'

        if neutral_only:
            conserved = b_values_input['Conserved']
            is_bytes = conserved.dtype.kind == 'S'
            neutral_mask = conserved == b'N' if is_bytes else conserved == 'N'

            x = positions[neutral_mask]
            y = b_vals[neutral_mask]
            sort_idx = np.argsort(x)
            x = x[sort_idx]
            y = y[sort_idx]

            diffs = np.diff(x)
            split_indices = np.where(diffs != 1)[0] + 1
            x_segments = np.split(x, split_indices)
            y_segments = np.split(y, split_indices)

            print(f"Plotting {len(x)} neutral positions in {len(x_segments)} segments.")
            for xs, ys in zip(x_segments, y_segments):
                if len(xs) > 1:
                    ax.plot(xs, ys, color='blue', lw=2, zorder=1, alpha=0.8, label=legend_name_blue)#"Calculated B")

            if len(x) > 0:
                ax.set_xlim(x.min() - 1, x.max())
        else:
            x = positions
            y = b_vals
            max_points = 10000
            if len(x) > max_points:
                idx = np.linspace(0, len(x) - 1, max_points).astype(int)
                x = x[idx]
                y = y[idx]
            ax.plot(x, y, color='blue', lw=2, zorder=1, alpha=0.8, label=legend_name_blue)#"Calculated B")
            ax.set_xlim(x.min() - 1, x.max())

    elif caller == "gene":
        x = b_values_input[:, 0]
        y = b_values_input[:, 1]
        ax.plot(x, y, color='blue', lw=2, zorder=1, alpha=0.8, label=legend_name_blue)#"Calculated B")
        ax.set_xlim(x.min() - 1, x.max())

    # Labels and title
    ax.set_ylabel('Expected diversity relative to neutral evolution (B)', fontsize=13)
    if caller == "chromosome":
        ax.set_title(f'B for chromosome {chrom} ({positions.min()}–{positions.max()} bp)', fontsize=15, fontweight='bold')
        if rec_rates is not None:
            ax.set_xlabel('Chromosomal position (bp)', fontsize=13, labelpad=40)
        else:
            ax.set_xlabel('Chromosomal position (bp)', fontsize=13)
    else:
        ax.set_xlabel('Distance from single selected element of size 10 kb', fontsize=13)
        ax.set_title(title_name, fontsize=15, fontweight='bold')

        # Plot B_uncorrected if provided
    if B_uncorrected is not None:
        uncorrected_data = load_B_uncorrected(B_uncorrected)
        if uncorrected_data is not None and len(uncorrected_data) > 0:
            ax.plot(
                uncorrected_data["Distance"],
                uncorrected_data["B"],
                color='black',
                lw=2,
                alpha=0.5,
                label=legend_name_orange
            )
        # Plot B_uncorrected if provided
    if B_observed is not None:
        observed_data = load_B_observed(B_observed)
        if observed_data is not None and len(observed_data) > 0:
            ax.scatter(
                observed_data["Distance"],
                observed_data["B"],
                color='#f57616',
                s=12,
                alpha=0.9,
                label=legend_name_dot#"Observed B (simulations)"
            )


    ax.tick_params(axis='both', which='major', labelsize=10)

    # Gene-range bars
    if gene_ranges is not None and len(gene_ranges) > 0:
        ymin, ymax = ax.get_ylim()
        bar_y = ymin - (ymax - ymin) * 0.05
        ax.set_ylim(bar_y, ymax)

        # convert start/end to int for plotting
        segments = [((int(start), bar_y), (int(end), bar_y)) for _, start, end in gene_ranges]
        ax.add_collection(LineCollection(segments, colors='black', linewidths=30))

        if caller == "chromosome" and not neutral_only:
            gene_mask = np.zeros_like(x, dtype=bool)
            for _, start, end in gene_ranges:
                start = int(start)
                end = int(end)
                gene_mask |= (x >= start) & (x <= end)
            idx = np.where(gene_mask)[0]
            splits = np.where(np.diff(idx) != 1)[0] + 1
            for seg in np.split(idx, splits):
                if len(seg) > 1:
                    coords = np.column_stack((x[seg], y[seg]))
                    ax.add_collection(LineCollection([coords], colors='black', linewidths=1.5))

    # X-axis formatting
    ax.xaxis.set_major_formatter(
        ticker.FuncFormatter(lambda value, pos:
            f"{int(value)} bp" if value < 1e3 else (
                f"{value/1e6:.2f} Mb" if value >= 1e6 else f"{int(value/1e3)} kb"
            )
        )
    )

    # Recombination rate strip
    if rec_rates is not None and caller == "chromosome":
        ax_rec = fig.add_subplot(gs[1], sharex=ax)
        ax_rec.set_yticks([])
        ax_rec.tick_params(axis='x', which='major', labelsize=9)
        rec_img = np.expand_dims(rec_rates, axis=0)
        min_pos = positions.min()
        extent = [min_pos, min_pos + len(rec_rates) * 20000, 0, 1]
        ax_rec.imshow(rec_img, aspect='auto', extent=extent, cmap='Purples', origin='lower')
        ax_rec.set_frame_on(False)
        plt.setp(ax.get_xticklabels(), visible=False)

    # Final layout and save
    if rec_rates is None:
        plt.tight_layout()
    else:
        fig.subplots_adjust(hspace=0.01, bottom=0.12)
    ax.legend(loc="lower right", fontsize=10, frameon=True)


    plt.savefig(output_path, dpi=300)
    print(f"Plot saved to {output_path}")

import numpy as np

def load_B_uncorrected(file_path):
    """
    Loads B_uncorrected data from a CSV file with header 'Distance,B'
    
    Parameters
    ----------
    file_path : str
        Path to the uncorrected B-values file

    Returns
    -------
    np.ndarray
        Structured array with fields 'Distance' and 'B'
    """
    try:
        data = np.genfromtxt(file_path, delimiter=",", names=True, dtype=[("Distance", float), ("B", float)])
        return data
    except Exception as e:
        print(f"Error reading B_observed file: {file_path}")
        raise e

def load_B_observed(file_path):
    """
    Loads B_observed data from a CSV file with header 'replicate,start,pi',
    averages across replicates by 'start', subtracts 9999 from 'start', and returns
    a structured array with fields 'Distance' and 'B'.
    
    Parameters
    ----------
    file_path : str
        Path to the observed B-values file
    
    Returns
    -------
    np.ndarray
        Structured array with fields 'Distance' and 'B'
    """
    try:
        raw = np.genfromtxt(file_path, delimiter=",", names=True)
        # Convert to Pandas for groupby convenience
        import pandas as pd
        df = pd.DataFrame(raw)
        grouped = df.groupby("start", as_index=False)["pi"].mean()
        grouped["Distance"] = grouped["start"] - 9999
        grouped["pi"] = grouped["pi"] / 0.012 #/ 0.00631579 # for selfing, / 0.00631579, else 0.012
        grouped.rename(columns={"pi": "B"}, inplace=True)
        result = grouped[["Distance", "B"]].to_records(index=False)
        return result
    except Exception as e:
        print(f"Error reading B_uncorrected file: {file_path}")
        raise e


#This script is to get the slope of recovery of pi and the number of  base pairs that would lead to 50%, 75% and 90% recovery for a given recombination rate.
#I'm here multiplying both the recombination and gene conversion rate by 0.5, just to see more of an effect of demography.

import sys
import math
import numpy as np

#Define variables and constants:
out_folder="droso_single_exon_gc_10kb_decline10x"
g = 0.5*1e-8 #1e-8 #rate of gene conversion
tract_len=440 #mean tract length of gene conversion in base pairs
r = 0.5*1.0*1e-8 #rate of recombination
l = 10000.0 #(*Length of genomic element*)
u = 3.0*1e-9 #(*Mutation rate*)
U = l*u
#Parameters of instantaneous change in demographic history:
Nanc = 1e6 #(Ancestral population size)
Ncur = 1e5 #(Current population size)
TIME="T_1" #T_0_1/T_0_5/T_1
time_of_change=1.0 #0.1/0.5/1(This is the time of change in 2Ncur generations in the past.)
#Parameters of the DFE:
DFE="DFE3"
f0 = 0.1 #(*Proportion of effectively neutral mutations with 0 <= |2Nes| < 1 *)
f1 = 0.1 #(*Proportion of weakly deleterious mutations with 1 <= |2Nes| < 10 *)
f2 = 0.1 #(*Proportion of moderately deleterious mutations with 10 <= |2Nes| < 100 *)
f3 = 0.7 #(*Proportion of strongly deleterious mutations with |2Nes| >= 100 *)
#(*Note that the number of classes can easily be increased to whatever is required to approximate the continuous DFE *)
h = 0.5 #(* dominance coefficient *)
gamma_cutoff = 2.0 #5.0
s_window_size = 100

#Constants that we do not need from the user:
pi = 4*Nanc*u #(*Expected nucleotide diversity under neutrality*)
#(*Now we define the boundaries of the fixed intervals over which we will integrate. The number of bins can be a user parameter, if we like. *)
t0 = 0.0
t1 = h*(1/(2*Nanc))
t1half = h*(gamma_cutoff/(2*Nanc)) #(* This is the cut-off value of 2Nes=5. This derivation assumes that all mutations with 2Nes<5 will not contribute to BGS *)
t2 = h*(10/(2*Nanc))
t3 = h*(100/(2*Nanc))
t4 = h*1.0

#calculate the quantities "a" and "b" which are constants that depend on the recombination and gene conversion rate and also the distance between the focal site and the functional element.
def calculate_a_and_b(posn):
    C = (1.0 - math.exp(-2.0*r*posn))/2.0
    if g==0:
        a = C
        b = C + r*l
    elif g > 0:
        if posn+l < 0.5*tract_len:#The 0.5 is currently aribtrary. It's just about when the approximation of 1-exp(-x)~x holds. 
            #print ("accounting for small-distance gene conversion")
            a = (C + (g*posn))
            b = (C + r*l + (g*(posn+l)))
        else:
            #print ("accounting for large-distance gene conversion")
            a = g*tract_len + C
            b = g*tract_len + r*l + C
    return (a, b)

#calculate the exponent using previously computed values of "a" and "b"
def calculate_exponent(t_start, t_end, posn):
    t_tmp = calculate_a_and_b(posn)
    a = t_tmp[0]
    b = t_tmp[1]
    E1 = ((U*a)/((1-a)*(a-b)*(t_end-t_start))) * math.log((a+(t_end*(1-a)))/(a + (t_start*(1-a))))
    E2 = -1.0*((U*b)/((1-b)*(a-b)*(t_end-t_start)))*math.log((b + ((1-b)*t_end))/(b + ((1-b)*t_start)))
    E = E1 + E2
    return (E)

#Calculate B due to a single functional element at the focal site. Here we sum over the DFE.
def calculate_B(posn):
    E_f1 = calculate_exponent(t1half, t2, posn) 
    E_f2 = calculate_exponent(t2, t3, posn)
    E_f3 = calculate_exponent(t3, t4, posn)
    E_bar = f0*0.0 + f1*((t1half-t1)/(t2-t1))*0.0 + f1*((t2-t1half)/(t2-t1))*E_f1 + f2*E_f2 + f3*E_f3
    B = math.exp(-1.0*E_bar)
    return (B)

print(calculate_B)





## Population genetic parameters for the simulated or empirical population
## Accurate estimation requires accurate and appropriate parameters
##
## e.g. Bvalcalc --params path/to/ExampleParams.py
##
## Core parameters
x = 1 # Scaling factor (N,u,r), keep as 1 unless calculating for rescaled simulations
Nanc = 1e6 / x # Ancestral population size [1]
r = 1e-8 * x # Recombination (crossover) rate per bp, per generation (sex-averaged) [2]
u = 3e-9 * x # Mutation rate (all types) per bp, per generation [1]
g = 1e-8 * x # Gene conversion initiation rate per bp, per generation [3]
k = 440 # Gene conversion tract length (bp) [3]
## DFE parameters for ALL sites in annotated regions (Sum must equal 1)
f0 = 0.25 # Proportion of effectively neutral mutations with 0 <= |2Ns| < 1 (Note that 2Ns<5 does not contribute to BGS) [4]
f1 = 0.49 # Proportion of weakly deleterious mutations with 1 <= |2Ns| < 10 [4]
f2 = 0.04 # Proportion of moderately deleterious mutations with 10 <= |2Ns| < 100 [4]
f3 = 0.22 # Proportion of strongly deleterious mutations with |2Ns| >= 100 [4]
## Demography parameters
Ncur = 2 * Nanc # Current population size (!Requires --pop_change) [5]
time_of_change = 0.45 # Time in Nanc generations ago that effective population size went from Nanc to Ncur (!Requires --pop_change) [6]
## Advanced DFE parameters 
h = 0.5 # Dominance coefficient of selected alleles [Naive value]
mean, shape, proportion_synonymous = 500, 0.5, 0.3 # Gamma distribution of DFE to discretize and replace f0-f3 [mean (2Ns), shape, proportion synonymous] (!Requires --gamma_dfe) [Naive value]## Literature cited
# [1] Keightley et al 2014  doi: 10.1534/genetics.113.158758
# [2] Comeron et al 2012 doi: 10.1371/journal.pgen.1002905
# [3] Miller et al 2016 doi: 10.1534/genetics.115.186486
# [4] Johri et al 2020 doi: 10.1534/genetics.119.303002
# [5] Laurent et al 2011 doi: 10.1093/molbev/msr031
# [6] Kapopoulou et al 2018 doi: 10.1093/gbe/evy185


from Bvalcalc.core.chromBcalc import chromBcalc
from Bvalcalc.utils.load_bed_gff import load_bed_gff
from Bvalcalc.utils.load_Bmap import load_Bmap
from Bvalcalc.core.calculateB import calculateB_unlinked
import numpy as np

def regionBcalc(args, region):    

    allblockstart, allblockend, allblockchrom  = load_bed_gff(args.bedgff) # Read BED/GFF, return start and end of conserved elements

    import Bvalcalc.utils.dfe_helper as dfe_helper
    dfe_helper.GAMMA_DFE = args.gamma_dfe # Update DFE if --gamma_dfe
    dfe_helper.CONSTANT_DFE = args.constant_dfe # Update DFE if --constant_dfe

    calc_chrom, calc_start, calc_end = parse_region(region)

    mask = allblockchrom == calc_chrom
    blockstart, blockend = allblockstart[mask], allblockend[mask]
    chromosome = calc_chrom
    unlinked_blockstart, unlinked_blockend = allblockstart[~mask], allblockend[~mask]
    unlinked_L = np.sum(unlinked_blockend-unlinked_blockstart)
    unlinked_B = calculateB_unlinked(unlinked_L)

    if args.prior_Bmap is not None:
        prior_chromosomes, prior_positions, prior_b_values = load_Bmap(file_path = args.prior_Bmap)
        if not args.quiet: print(f"Using prior B values from {args.prior_Bmap}")
        prior_mask = (prior_chromosomes == chromosome)
        prior_pos = prior_positions[prior_mask]
        prior_b = prior_b_values[prior_mask]
    else:
        prior_pos, prior_b = None, None

    if args.out is not None: # Overwrite existing file with header
        with open(args.out, 'w') as out_f:
            out_f.write("Chromosome,Start,B\n")
    output_data, block_ranges, rec_rate_per_chunk_in_region, chunk_size = chromBcalc(args, blockstart, blockend, chromosome, unlinked_B, prior_pos, prior_b, calc_start, calc_end, caller="regionBcalc")

    return  output_data, block_ranges, rec_rate_per_chunk_in_region, chunk_size

def parse_region(region_str):
    try:
        chrom_part, pos_part = region_str.split(":")
        start_str, end_str = pos_part.split("-")
        chrom = chrom_part
        start = int(start_str.replace(",", ""))
        end = int(end_str.replace(",", ""))
        return chrom, start, end
    except ValueError:
        raise ValueError(f"Region format invalid: '{region_str}' should be like 'chr1:12345-67890'")


import numpy as np
import sys

def write_chrom_B_to_file(out,
                          output_data,
                          quiet,
                          hri_starts=None,
                          hri_ends=None,
                          binsize=None,
                          calc_end=None):
    """
    Write Chromosome,Start,B to CSV.
    If hri_starts/hri_ends provided, append ' to B for rows whose bin overlaps any HRI region.
    - output_data: structured array with fields ['Chromosome','Start','B']
    - hri_starts, hri_ends: 1D int arrays of equal length (inclusive coordinates)
    - binsize: int, bin length (last bin truncated)
    - calc_end: int, max coordinate to cap last-bin end (optional but recommended)
    """

    # Fast path: no HRI spans or no binsize provided -> write floats directly
    need_mark = (
        hri_starts is not None and hri_ends is not None and
        len(hri_starts) > 0 and len(hri_ends) > 0 and
        binsize is not None
    )

    if not need_mark:
        with open(out, 'a') as f:
            np.savetxt(f, output_data, delimiter=",", fmt="%s,%d,%.6f", comments="")
        return

    # Build bin start/end coordinates
    bin_starts = output_data['Start'].astype(np.int64)
    if calc_end is None:
        # Fallback: cap ends at start + binsize - 1
        bin_ends = bin_starts + (int(binsize) - 1)
    else:
        bin_ends = np.minimum(bin_starts + (int(binsize) - 1), int(calc_end)).astype(np.int64)

    # Sort HRI intervals for binary searches
    s = np.sort(np.asarray(hri_starts, dtype=np.int64))
    e = np.sort(np.asarray(hri_ends,   dtype=np.int64))

    # Vectorized interval overlap test:
    # overlap if count(starts <= bin_end) - count(ends < bin_start) > 0
    starts_le_end = np.searchsorted(s, bin_ends, side='right')
    ends_lt_start = np.searchsorted(e, bin_starts - 1, side='right')
    overlaps = (starts_le_end - ends_lt_start) > 0

    # Prepare per-row B strings with optional '*'
    B = output_data['B'].astype(float)
    B_as_str = np.array([f"{v:.6f}" for v in B], dtype="<U16")
    if overlaps.any():
        B_as_str[overlaps] = np.char.add(B_as_str[overlaps], "'")

    # Compose rows (don’t mutate output_data dtype)
    rows = np.empty((output_data.shape[0], 3), dtype=object)
    rows[:, 0] = output_data['Chromosome']
    rows[:, 1] = output_data['Start'].astype(int)
    rows[:, 2] = B_as_str

    if not quiet and hri_starts is not None: 
        print(f"B values calculated for HRI regions (B') is indicated with ' at end of line")

    with open(out, 'a') as f:
        np.savetxt(f, rows, delimiter=",", fmt="%s,%d,%s", comments="")




"""
bvalcalc: calculate relative diversity (B) under background selection.
"""

__version__ = "0.7.2"

# Expose main entry point
from .cli import main
from .core.calculateB import calculateB_linear, calculateB_recmap, calculateB_unlinked, get_params


__all__ = [
    "get_params", "calculateB_linear", "calculateB_unlinked",
    "main",
    "__version__",
]


import scipy.stats as st
import os
import importlib.util
from typing import Dict, Any
from functools import lru_cache

GAMMA_DFE = False  # Default, instead of prop injected
CONSTANT_DFE = False # Default, instead of prop injected

def get_DFE_params(params_path: str | None = None, gamma_dfe: bool = False, constant_dfe: bool = False) -> Dict[str, Any]:
    """
    Load and validate population parameters from the file pointed to by
    `params_path` or, if None, by the BCALC_params env var.
    Returns a dictionary of parameters for use in B-value calculations.
    """
    # 1. Determine the path: either passed in or from the env var
    if params_path is None:
        params_path = os.environ.get("BCALC_params")
        if not params_path:
            raise KeyError(
                "Environment variable BCALC_params not set. "
                "Cannot load pop-gen parameters."
            )

    # 2. Load the module
    spec = importlib.util.spec_from_file_location("params", params_path)
    if spec is None or spec.loader is None:
        raise FileNotFoundError(f"Could not load spec for params from {params_path}")
    pop = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(pop)

    # 3. Extract and validate required attributes
    required_names = ["g", "k", "r", "u", "Nanc", "h", "f0", "f1", "f2", "f3"]
    params: Dict[str, Any] = {}
    for name in required_names:
        if not hasattr(pop, name):
            raise AttributeError(f"params must define '{name}'")
        val = getattr(pop, name)
        if val is None:
            raise AttributeError(f"params parameter '{name}' is None; must be a numeric value")
        params[name] = float(val)

    # 4. Optional gamma-DFE override
    if GAMMA_DFE or gamma_dfe is not False: # The GAMMA_DFE is prop injected by CLI, gamma_dfe is provided by API
        mean = getattr(pop, 'mean', None)
        shape = getattr(pop, 'shape', None)
        prop_syn = getattr(pop, 'proportion_synonymous', None)
        if mean is None or shape is None or prop_syn is None:
            raise AttributeError(
                "params must define 'mean', 'shape' and 'proportion_synonymous' when --gamma_dfe is active"
            )
        from .dfe_helper import gammaDFE_to_discretized
        f0, f1, f2, f3 = gammaDFE_to_discretized(mean, shape, prop_syn)
        params.update({"f0": f0, "f1": f1, "f2": f2, "f3": f3})


    # 5. Set derived parameters and thresholds
    params["gamma_cutoff"] = 5
    params["t0"] = 0.0
    Nanc = params["Nanc"]
    h = params["h"]
    # Calculate generation-scale thresholds
    params["t1"] = h * (1.0 / (2.0 * Nanc))
    params["t1half"] = h * (params["gamma_cutoff"] / (2.0 * Nanc))
    params["t2"] = h * (10.0 / (2.0 * Nanc))
    params["t3"] = h * (100.0 / (2.0 * Nanc))
    params["t4"] = h * 1.0

    if CONSTANT_DFE or constant_dfe is not False: # The CONSTANT_DFE is prop injected by CLI, constant_dfe is provided by API
        s = getattr(pop, "s", None)
        prop_syn = getattr(pop, 'proportion_synonymous', None)
        if s is None or prop_syn is None:
            raise AttributeError(
                "params must define 's' and 'proportion_synonymous' when --constant_dfe is active"
            )
        params["t_constant"] = h * s # Set parameter to be exported to calculateB
    else:         
        params["t_constant"] = None

    return params


def gammaDFE_to_discretized(mean: float, shape: float, proportion_synonymous: float):
    if mean <= 0 or shape <= 0:
        raise ValueError("`mean` and `shape` must be positive.")
    if not (0 <= proportion_synonymous < 1):
        raise ValueError("`proportion_synonymous` must be in [0, 1).")

    theta = mean / shape               # scale parameter
    dist  = st.gamma(a=shape, scale=theta)

    # cumulative probabilities at the cut‑points
    c1   = dist.cdf(1.0)
    c10  = dist.cdf(10.0)
    c100 = dist.cdf(100.0)


    f0 = c1                      # (0, 1]
    f1 = c10  - c1               # (1,10]
    f2 = c100 - c10              # (10,100]
    f3 = 1.0   - c100            # >100

    # 4) scale to sum to (1 - p_syn)
    scale = 1.0 - proportion_synonymous
    f0, f1, f2, f3 = (f0 * scale,
                      f1 * scale,
                      f2 * scale,
                      f3 * scale)

    # 5) add synonymous fraction back into f0
    f0 += proportion_synonymous

    print(f"Converting gamma distribution to discretized DFE")
    print(f"Gamma params: mean = {mean}, shape = {shape}, scale = {theta}")
    print(f"Inferred f0, f1, f2, f3 = ", f0, f1, f2, f3)

    return f0, f1, f2, f3


## Population genetic parameters for the simulated or empirical population
## Accurate estimation requires accurate and appropriate parameters
##
## e.g. Bvalcalc --params path/to/TemplateParams.py
##
## Core parameters
f = 0 # Selfing rate (f = S/(2-S); Wright's inbreeding coefficient) [Naive value]
x = 1 # Scaling factor (N,u,r), keep as 1 unless calculating for rescaled simulations
Nanc = 1e5 / (1+f) / x # Ancestral population size [Naive value]
r = 1e-8 * (1-f) * x # Recombination (crossover) rate per bp, per generation (sex-averaged) [Naive value]
u = 1e-8 * x # Mutation rate (all types) per bp, per generation [Naive value]
g = 1e-8 * (1-f) * x # Gene conversion initiation rate per bp, per generation [Naive value]
k = 500 # Gene conversion tract length (bp) [Naive value]
## DFE parameters for ALL sites in annotated regions (Sum must equal 1)
f0 = 0.25 # Proportion of effectively neutral mutations with 0 <= |2Ns| < 1 (Note that 2Ns<5 does not contribute to BGS) [Naive value]
f1 = 0.25 # Proportion of weakly deleterious mutations with 1 <= |2Ns| < 10 [Naive value]
f2 = 0.25 # Proportion of moderately deleterious mutations with 10 <= |2Ns| < 100 [Naive value]
f3 = 0.25 # Proportion of strongly deleterious mutations with |2Ns| >= 100 [Naive value]
# ## Demography parameters
Ncur = Nanc # Current population size (!Requires --pop_change) [Naive value] 
time_of_change = 1 # Time in Nanc generations ago that effective population size went from Nanc to Ncur (!Requires --pop_change) [Naive value]
## Advanced DFE parameters 
h = 0.5 + (f-0.5*f) # Dominance coefficient of selected alleles [Naive value]
mean, shape, proportion_synonymous = 100, 1, 0.3 # Gamma distribution of DFE to discretize and replace f0-f3 [mean (2Ns), shape, proportion synonymous] (!Requires --gamma_dfe) [Naive value]
## Literature cited
# [1]
# [2]
# [3]
# [4]



import numpy as np
import sys

def bin_outputs(b_values: np.ndarray,
                positions: np.ndarray,
                binsize: int
               ) -> tuple[np.ndarray, np.ndarray]:
    """
    Collapse b_values & positions into consecutive bins of length `binsize`.
    Returns
    -------
    binned_b_values : float array, shape=(ceil(N/binsize),)
        Mean of b_values in each bin.
    binned_positions : int array, same length as binned_b_values
        The first position in each bin (i.e. positions[0], positions[binsize], …).
    """
    if b_values.shape[0] != positions.shape[0]:
        print("Shapes", b_values.shape[0], positions.shape[0])
        raise ValueError("`b_values` and `positions` must be the same length")

    N = b_values.shape[0]
    # bin start indices: 0, binsize, 2*binsize, …
    bin_idx = np.arange(0, N, binsize, dtype=int)

    # fast C‐loop to sum each bin
    sums = np.add.reduceat(b_values, bin_idx)

    # how many items in each bin (last bin may be shorter)
    counts = np.diff(np.append(bin_idx, N))

    # compute mean
    binned_b_values = sums / counts

    # record the first genomic position in each bin
    binned_positions = positions[bin_idx]

    return binned_b_values, binned_positions



"""
Sample data utilities for Bvalcalc.
"""
import os
import shutil
import sys
from pathlib import Path


def get_sample_data_dir():
    """
    Get the directory where sample data should be installed.
    
    Returns:
        str: Path to the sample data directory
    """
    # Use current working directory as default
    return os.getcwd()


def get_package_data_dir():
    """
    Get the directory where sample data is stored in the package.
    
    Returns:
        str: Path to the package data directory
    """
    try:
        # Try to use importlib.resources for modern Python packaging
        import importlib.resources as resources
        with resources.path('Bvalcalc', 'data') as data_path:
            return str(data_path)
    except (ImportError, FileNotFoundError):
        # Fallback for older Python versions or development environment
        # Get the directory where this module is located
        current_dir = os.path.dirname(os.path.abspath(__file__))
        # Go up to Bvalcalc root, then to data directory
        package_root = os.path.dirname(os.path.dirname(current_dir))
        return os.path.join(package_root, 'data')


def download_sample_data(force=False, quiet=False, target_dir=None):
    """
    Download Drosophila CDS sample data to user-accessible location.
    
    Args:
        force (bool): If True, overwrite existing data
        quiet (bool): If True, suppress output messages
        target_dir (str): Target directory (defaults to current working directory)
    
    Returns:
        bool: True if successful, False otherwise
    """
    source_dir = get_package_data_dir()
    if target_dir is None:
        target_dir = get_sample_data_dir()
    
    if not os.path.exists(source_dir):
        if not quiet:
            print(f"Error: Sample data source directory not found: {source_dir}")
        return False
    
    # Create target directory if it doesn't exist
    os.makedirs(target_dir, exist_ok=True)
    
    if not quiet:
        print(f"Downloading Drosophila CDS sample data to: {target_dir}")
    
    # Define which files to copy (Drosophila CDS only)
    files_to_copy = [
        'cds_noX.bed',           # CDS annotations
        'dmel_comeron_recmap.csv', # Drosophila recombination map
        'README.md'              # Documentation
    ]
    
    try:
        existing_files = []
        copied_files = []
        
        # Copy only the specified files
        for filename in files_to_copy:
            source_path = os.path.join(source_dir, filename)
            target_path = os.path.join(target_dir, filename)
            
            if not os.path.exists(source_path):
                if not quiet:
                    print(f"Warning: Source file not found: {filename}")
                continue
            
            if os.path.exists(target_path) and not force:
                existing_files.append(filename)
                continue
            
            shutil.copy2(source_path, target_path)
            copied_files.append(filename)
        
        # Show warnings for existing files
        if existing_files and not quiet:
            print(f"Warning: The following files already exist and were not overwritten:")
            for item in existing_files:
                print(f"  {item}")
            print()
        
        # Show copied files
        if copied_files and not quiet:
            for item in copied_files:
                print(f"  Copied {item}")
        
        if not quiet:
            print(f"\nDrosophila CDS sample data successfully downloaded!")
            print(f"Available files:")
            for item in sorted(os.listdir(target_dir)):
                item_path = os.path.join(target_dir, item)
                if os.path.isfile(item_path):
                    size = os.path.getsize(item_path)
                    print(f"  {item} ({size:,} bytes)")
                else:
                    print(f"  {item}/")
            print(f"\nYou can now use these files with Bvalcalc commands.")
            print(f"Example: bvalcalc --genome --params your_params.py --bedgff cds_noX.bed --rec_map dmel_comeron_recmap.csv")
        
        return True
        
    except Exception as e:
        if not quiet:
            print(f"Error downloading sample data: {e}")
        return False


def list_sample_data(quiet=False):
    """
    List available Drosophila CDS sample data files.
    
    Args:
        quiet (bool): If True, suppress output messages
    
    Returns:
        bool: True if successful, False otherwise
    """
    data_dir = get_sample_data_dir()
    
    if not os.path.exists(data_dir):
        if not quiet:
            print("Drosophila CDS sample data not found. Run 'bvalcalc --download_sample_data' to download sample files.")
        return False
    
    if not quiet:
        print(f"Drosophila CDS sample data available in: {data_dir}")
        print("Available files:")
        
        for item in sorted(os.listdir(data_dir)):
            item_path = os.path.join(data_dir, item)
            if os.path.isfile(item_path):
                size = os.path.getsize(item_path)
                print(f"  {item} ({size:,} bytes)")
            else:
                print(f"  {item}/")
    
    return True



import numpy as np

def load_vcf(vcf_path):
    print(f"loading VCF/CSV: {vcf_path}")
    chromosomes = []
    positions   = []

    with open(vcf_path, 'r') as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith('#'):
                continue

            # Detect comma-separated two-column CSV vs. whitespace-separated VCF
            if ',' in line:
                parts = [p.strip() for p in line.split(',')]
                # only accept exactly two columns
                if len(parts) != 2:
                    continue
            else:
                parts = line.split()
                # need at least chromosome and POS
                if len(parts) < 2:
                    continue

            chrom = parts[0]
            try:
                pos = int(parts[1])
            except ValueError:
                # bad POS field, skip
                continue

            chromosomes.append(chrom)
            positions.append(pos)

    chrom_array = np.array(chromosomes, dtype='<U20')
    pos_array   = np.array(positions,   dtype=np.int64)
    return chrom_array, pos_array



from Bvalcalc.core.calculateB import calculateB_linear, calculateB_recmap
from Bvalcalc.core.helpers.calc_R_len_dist import calc_R_lendist_for_chunks
import numpy as np

def calc_B_from_chunks(chunk_idx, chunk_size, chr_start, chr_size, num_chunks, 
                    precise_chunks, lperchunk, rec_rate_per_chunk, gc_rate_per_chunk, excluded_chunks = None):

    if excluded_chunks is not None:
        lperchunk_filtered = lperchunk.copy() # Branch lperchunk_filtered from lperchunk
        lperchunk_filtered[np.asarray(excluded_chunks, dtype=int)] = 0 # Set lperchunk to 0 for excluded chunks
    else: lperchunk_filtered = lperchunk # Else it's just lperchunk

    chunk_starts = chr_start + np.arange(num_chunks) * chunk_size
    chunk_ends = np.minimum(chunk_starts + chunk_size - 1, chr_size)
    chunk_mids = (chunk_ends + chunk_starts) / 2

    chunk_pseudoblockstart = chunk_mids - 0.5 * lperchunk_filtered # Pseudoblocks are the combined selected regions in each chunk, which are combined into a single central "pseudoblock" aka single selected region
    chunk_pseudoblockend = chunk_mids + 0.5 * lperchunk_filtered

    upstream_pseudochunk_mask = np.ones(num_chunks, dtype=bool)
    downstream_pseudochunk_mask = np.ones(num_chunks, dtype=bool)

    upstream_pseudochunk_mask[max(0, chunk_idx - precise_chunks):] = False # Mask for downstream blocks
    upstream_pseudochunk_mask[lperchunk_filtered==0] = False
    downstream_pseudochunk_mask[0:min(num_chunks, chunk_idx + precise_chunks + 1)] = False # Mask for upstream blocks
    downstream_pseudochunk_mask[lperchunk_filtered==0] = False
    
    relevant_upstream_psdc_lengths = lperchunk_filtered[upstream_pseudochunk_mask] # Use mask on pseudochunk lengths
    relevant_downstream_psdc_lengths = lperchunk_filtered[downstream_pseudochunk_mask] # Use mask on pseudochunk lengths

    relevant_upstream_pseudoblockends = chunk_pseudoblockend[upstream_pseudochunk_mask]
    relevant_downstream_pseudoblockstarts = chunk_pseudoblockstart[downstream_pseudochunk_mask]
    relevant_upstream_psdc_distances = chunk_mids[chunk_idx] - relevant_upstream_pseudoblockends - 1
    relevant_downstream_psdc_distances = relevant_downstream_pseudoblockstarts - chunk_mids[chunk_idx] - 1

    if rec_rate_per_chunk is not None: # IF REC_RATE MAP IS AVAILABLE 
        # Get the indices for upstream and downstream pseudochunks
        chunk_rec_distances = (chunk_ends - chunk_starts + 1) * rec_rate_per_chunk
        upstream_indices = np.nonzero(upstream_pseudochunk_mask)[0]
        downstream_indices = np.nonzero(downstream_pseudochunk_mask)[0]

        upstream_rec_lengths, downstream_rec_lengths, upstream_rec_distances, downstream_rec_distances = calc_R_lendist_for_chunks(
            upstream_indices, downstream_indices, rec_rate_per_chunk, 
            relevant_upstream_psdc_lengths, relevant_downstream_psdc_lengths, 
            chunk_idx, chunk_size, relevant_upstream_pseudoblockends, relevant_downstream_pseudoblockstarts, 
            chunk_starts, chunk_ends, chunk_rec_distances, num_chunks
            ) # Get local r * lengths for length of, and distances to pseudoblocks for each chunk
        
    if gc_rate_per_chunk is not None: # IF GC_RATE MAP IS AVAILABLE 
        # Get the indices for upstream and downstream pseudochunks
        chunk_gc_distances = (chunk_ends - chunk_starts + 1) * gc_rate_per_chunk
        upstream_indices = np.nonzero(upstream_pseudochunk_mask)[0]
        downstream_indices = np.nonzero(downstream_pseudochunk_mask)[0]

        upstream_gc_lengths, downstream_gc_lengths, upstream_gc_distances, downstream_gc_distances = calc_R_lendist_for_chunks(
            upstream_indices, downstream_indices, gc_rate_per_chunk, 
            relevant_upstream_psdc_lengths, relevant_downstream_psdc_lengths, 
            chunk_idx, chunk_size, relevant_upstream_pseudoblockends, relevant_downstream_pseudoblockstarts, 
            chunk_starts, chunk_ends, chunk_gc_distances, num_chunks
            ) # Get local r * lengths for length of, and distances to pseudoblocks for each chunk


    #Run calculateBs
    if rec_rate_per_chunk is not None and gc_rate_per_chunk is not None: # IF REC_RATE MAP IS AVAILABLE and GC IS AVAILABLE
        relevant_upstream_psdc_B = np.prod(calculateB_recmap(distance_to_element=relevant_upstream_psdc_distances, length_of_element=relevant_upstream_psdc_lengths, 
                                                             rec_distances=upstream_rec_distances, rec_lengths=upstream_rec_lengths, 
                                                               gc_distances=upstream_gc_distances, gc_lengths=upstream_gc_lengths))
        relevant_downstream_psdc_B = np.prod(calculateB_recmap(distance_to_element=relevant_downstream_psdc_distances, length_of_element=relevant_downstream_psdc_lengths, 
                                                               rec_distances=downstream_rec_distances, rec_lengths=downstream_rec_lengths, 
                                                               gc_distances=downstream_gc_distances, gc_lengths=downstream_gc_lengths))
    elif rec_rate_per_chunk is not None and gc_rate_per_chunk is None: # IF REC_RATE MAP IS AVAILABLE and GC NOT AVAILABLE
        relevant_upstream_psdc_B = np.prod(calculateB_recmap(distance_to_element=relevant_upstream_psdc_distances, length_of_element=relevant_upstream_psdc_lengths, 
                                                             rec_distances=upstream_rec_distances, rec_lengths=upstream_rec_lengths, 
                                                               gc_distances=None, gc_lengths=None))
        relevant_downstream_psdc_B = np.prod(calculateB_recmap(distance_to_element=relevant_downstream_psdc_distances, length_of_element=relevant_downstream_psdc_lengths, 
                                                               rec_distances=downstream_rec_distances, rec_lengths=downstream_rec_lengths, 
                                                               gc_distances=None, gc_lengths=None))
    elif rec_rate_per_chunk is None and gc_rate_per_chunk is not None: # IF REC_RATE MAP NOT AVAILABLE and GC IS AVAILALBE
        relevant_upstream_psdc_B = np.prod(calculateB_recmap(distance_to_element=relevant_upstream_psdc_distances, length_of_element=relevant_upstream_psdc_lengths, 
                                                             rec_distances=None, rec_lengths=None, 
                                                               gc_distances=upstream_gc_distances, gc_lengths=upstream_gc_lengths))
        relevant_downstream_psdc_B = np.prod(calculateB_recmap(distance_to_element=relevant_downstream_psdc_distances, length_of_element=relevant_downstream_psdc_lengths, 
                                                               rec_distances=None, rec_lengths=None, 
                                                               gc_distances=downstream_gc_distances, gc_lengths=downstream_gc_lengths))
    else: # NEITHER REC_MAP NOR GC_MAP AVAILABLE
        relevant_upstream_psdc_B = np.prod(calculateB_linear(relevant_upstream_psdc_distances, relevant_upstream_psdc_lengths))
        relevant_downstream_psdc_B = np.prod(calculateB_linear(relevant_downstream_psdc_distances, relevant_downstream_psdc_lengths))

    return relevant_downstream_psdc_B * relevant_upstream_psdc_B # Return a single B value that applies to all sites in focal chunk


from Bvalcalc.core.helpers.demography_helpers import get_Bcur

def siteBcalc(args):    
    element_size, distance, quiet = args.element_size, args.distance, args.quiet

    import Bvalcalc.utils.dfe_helper as dfe_helper
    dfe_helper.GAMMA_DFE = args.gamma_dfe # Update DFE if --gamma_dfe
    dfe_helper.CONSTANT_DFE = args.constant_dfe # Update DFE if --constant_dfe
    from Bvalcalc.core.calculateB import calculateB_linear

    b_values = calculateB_linear(distance, element_size)
    print(f"B for site {distance}bp away from {element_size}bp region: {b_values}")

    if args.pop_change:
        if not quiet: print("Demographic change prior to B-calculation", b_values)
        b_values = get_Bcur(b_values)
        if not quiet: print("Demographic change applied to B-calculation", b_values)
    
    return


import numpy as np
from Bvalcalc.core.calculateB import calculateB_linear

def test_B_value_for_zero_length():
    result = calculateB_linear(np.array([100]), np.array([0]))
    assert np.allclose(result, 1.0), "B should be 1.0 when length_of_element is zero"


import subprocess
import sys
import re
from pathlib import Path

# Invoke the CLI via the installed package entry point.
BASE_CMD = [sys.executable, "-m", "Bvalcalc.cli"]

def test_cli_site_basic():
    # poetry run Bvalcalc --site --params tests/testparams/gcBasicParams.py --distance 100 --element_size 5000
    params = Path(__file__).parent / "testparams" / "gcBasicParams.py"
    cmd = BASE_CMD + [
        "--site",
        "--params", str(params),
        "--distance", "100",
        "--element_size", "5000",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)

    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    assert "B for site 100bp away from 5000bp region:" in result.stdout
    assert "0.9246145075198539" in result.stdout

def test_cli_gene_basic():
    # poetry run Bvalcalc --gene --params tests/testparams/nogcBasicParams.py --element_size 10000
    params = Path(__file__).parents[1] / "tests" / "testparams" / "nogcBasicParams.py"
    cmd = BASE_CMD + [
        "--gene",
        "--params", str(params),
        "--element_size", "10000",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    assert "====== R E S U L T S ! =============================" in out
    assert "B for adjacent site: 0.8049242606161049" in out
    assert "Mean B for flanking region: 0.9761402805820517" in out
    assert "No output CSV requested; skipping save." in out
    assert "= B value calculated" in out

def test_cli_gene_gcparams(tmp_path):
    # poetry run Bvalcalc --gene --params tests/testparams/gcBasicParams.py --element_size 10000 --plot tests/testout/test_plot.png
    params   = Path(__file__).parents[1] / "tests" / "testparams" / "gcBasicParams.py"
    plot_path = Path(__file__).parents[1] / "tests" / "testout" / "test_plot.png"
    cmd = BASE_CMD + [
        "--gene",
        "--params", str(params),
        "--element_size", "10000",
        "--plot", str(plot_path),
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    assert "B for adjacent site: 0.8910346781386976" in out
    assert "Mean B for flanking region: 0.9810661565709757" in out
    assert "B at start and end of the neutral region" in out
    assert "====== P L O T T I N G . . . =======================" in out
    assert f"Plot saved to {plot_path}" in out
    assert plot_path.exists(), f"Expected plot at {plot_path}, but not found"

def test_cli_genome_basic(tmp_path):
    # poetry run Bvalcalc --genome --params tests/testparams/nogcBasicParams.py --bedgff tests/testfiles/200kb_slimtest.csv --chr_sizes tests/testfiles/test_sizes.txt
    # poetry run Bvalcalc --region chr_200kb:1-200000 --params tests/testparams/nogcBasicParams.py --bedgff tests/testfiles/200kb_slimtest.csv --plot
    params         = Path(__file__).parents[1] / "tests" / "testparams" / "nogcBasicParams.py"
    bed_path       = Path(__file__).parents[1] / "tests" / "testfiles" / "200kb_slimtest.csv"
    chr_sizes_path = Path(__file__).parents[1] / "tests" / "testfiles" / "test_sizes.txt"
    output_path    = tmp_path / "200kb_dfe5.bvals"
    cmd = BASE_CMD + [
        "--genome",
        "--params", str(params),
        "--bedgff", str(bed_path),
        "--chr_sizes", str(chr_sizes_path),
        "--out", str(output_path),
        "--out_binsize", "1",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    match = re.search(r"Mean B of neutral sites across chromosome chr_200kb: ([0-9.]+)", out)
    assert match, "Could not find mean B output in CLI output"
    mean_b = float(match.group(1))
    expected = 0.753693843332108
    assert abs(mean_b - expected) < 1e-10, f"Expected {expected}, got {mean_b}"
    assert output_path.exists(), "Expected output file not created"
    assert output_path.stat().st_size > 0, "Output file is empty"

def test_cli_genome_gcparams(tmp_path):
    # poetry run Bvalcalc --genome --params tests/testparams/gcBasicParams.py --bedgff tests/testfiles/200kb_slimtest.csv --chr_sizes tests/testfiles/test_sizes.txt --out <tmp>/gc_bvals.bvals --out_binsize 1
    params         = Path(__file__).parents[1] / "tests" / "testparams" / "gcBasicParams.py"
    bed_path       = Path(__file__).parents[1] / "tests" / "testfiles" / "200kb_slimtest.csv"
    chr_sizes_path = Path(__file__).parents[1] / "tests" / "testfiles" / "test_sizes.txt"
    output_path    = tmp_path / "gc_bvals.bvals"
    cmd = BASE_CMD + [
        "--genome",
        "--params", str(params),
        "--bedgff", str(bed_path),
        "--chr_sizes", str(chr_sizes_path),
        "--out", str(output_path),
        "--out_binsize", "1",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    assert "====== R E S U L T S ====== S U M M A R Y ==========" in out
    assert "Mean B of neutral sites across chromosome chr_200kb: 0.836347850423207" in out
    assert output_path.exists(), "Expected output file not created"
    assert output_path.stat().st_size > 0, "Output file is empty"

def test_cli_genome_with_recmap_plot(tmp_path):
    # poetry run Bvalcalc --genome --params tests/testparams/nogcBasicParams.py --bedgff tests/testfiles/200kb_slimtest.csv --chr_sizes tests/testfiles/test_sizes.txt --rec_map tests/testfiles/200kb.map --out examples/200kb_dfe5.bvals --out_binsize 1
    params         = Path(__file__).parents[1] / "tests" / "testparams" / "nogcBasicParams.py"
    bed_path       = Path(__file__).parents[1] / "tests" / "testfiles" / "200kb_slimtest.csv"
    map_path       = Path(__file__).parents[1] / "tests" / "testfiles" / "200kb.map"
    chr_sizes_path = Path(__file__).parents[1] / "tests" / "testfiles" / "test_sizes.txt"
    output_path    = tmp_path / "200kb_dfe5.bvals"
    cmd = BASE_CMD + [
        "--genome",
        "--params", str(params),
        "--bedgff", str(bed_path),
        "--chr_sizes", str(chr_sizes_path),
        "--rec_map", str(map_path),
        "--out", str(output_path),
        "--out_binsize", "1",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    assert "Cumulative length of chromosome under selection: 99990bp (50.0%)" in out
    match = re.search(r"Mean B of neutral sites across chromosome chr_200kb: ([0-9.]+)", out)
    assert match, "Could not find mean B output in CLI output"
    mean_b = float(match.group(1))
    expected = 0.701584724570370
    assert abs(mean_b - expected) < 1e-10, f"Expected {expected}, got {mean_b}"
    assert f"Appended B values to: {output_path.as_posix()}" in out
    assert output_path.exists(), "Expected output file not created"
    assert output_path.stat().st_size > 0, "Output file is empty"

def test_cli_mean_b_value(tmp_path):
    # poetry run Bvalcalc --region chr_200kb:1514-62456 --params tests/testparams/nogcBasicParams.py --bedgff tests/testfiles/200kb_slimtest.csv --plot tests/testout/genome_test.png 
    params = "tests/testparams/nogcBasicParams.py"
    cmd = BASE_CMD + [
        "--region", "chr_200kb:1514-62456",
        "--params", params,
        "--bedgff", "tests/testfiles/200kb_slimtest.csv",
        "--plot", "tests/testout/genome_test.png",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    assert result.returncode == 0, f"Process failed: {result.stderr}"
    assert "Plot saved to tests/testout/genome_test.png" in out
    match = re.search(r"Mean B of neutral sites across specified region:\s+([0-9.]+)", result.stdout)
    assert match, "Could not find mean B output in CLI output"
    mean_b = float(match.group(1))
    expected = 0.7609515711751818
    assert abs(mean_b - expected) < 1e-10, f"Expected {expected}, got {mean_b}"

def test_cli_gene_contract():
    # poetry run Bvalcalc --gene --params tests/testparams/ContractParams_5N_1T.py --pop_change --plot
    params = Path(__file__).parents[1] / "tests" / "testparams" / "ContractParams_5N_1T.py"
    cmd = BASE_CMD + [
        "--gene",
        "--params", str(params),
        "--pop_change",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    assert "Mean B for flanking region: 0.9623874578845304" in out

def test_cli_gene_expand():
    # poetry run Bvalcalc --gene --params tests/testparams/ExpandParams_5N_1T.py --pop_change
    params = Path(__file__).parents[1] / "tests" / "testparams" / "ExpandParams_5N_1T.py"
    cmd = BASE_CMD + [
        "--gene",
        "--params", str(params),
        "--pop_change",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    assert "Mean B for flanking region: 0.982376743223556" in out
    assert "B prior to demographic change" in out
    assert "B post B-calculation" in out

def test_cli_selfing():
    # poetry run Bvalcalc --gene --params tests/testparams/SelfParams_0.9S_0.5h.py
    params = Path(__file__).parents[1] / "tests" / "testparams" / "SelfParams_0.9S_0.5h.py"
    cmd = BASE_CMD + [
        "--gene",
        "--params", str(params),
        "--pop_change",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    assert "B for adjacent site: 0.6034770660828896" in out
    assert "Mean B for flanking region: 0.8043714716235398" in out
    assert "B at start and end of the neutral region: [0.60347707 0.6035028  0.60352854 ... 0.89001725 0.89001929 0.89002133]" in out

def test_cli_positions_minimum_filter():
    # poetry run Bvalcalc --Bmap ./tests/testfiles/false_Bvalues_chr3R.csv --positions ./tests/testfiles/posfile.csv --out_minimum 0.5
    bmap_path = Path(__file__).parents[1] / "tests" / "testfiles" / "false_Bvalues_chr3R.csv"
    pos_path  = Path(__file__).parents[1] / "tests" / "testfiles" / "posfile.csv"

    cmd = BASE_CMD + [
        "--Bmap", str(bmap_path),
        "--positions", str(pos_path),
        "--out_minimum", "0.5",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    assert "Mean B across filtered sites: 0.562500" in out
    assert "Max B across filtered sites: 0.750000 at chr_2R:20000" in out
    assert "Min B across filtered sites: 0.500000 at chr_2R:1" in out

def test_cli_genome_hri_marking(tmp_path):
    # poetry run Bvalcalc --region chr_200kb:1-199999  --params ./tests/testparams/InterfParams.py  --bedgff ./tests/testfiles/200kb_hri.csv --rec_map ./tests/testfiles/200kb_hri.map --precise_chunks 2 --out haut.B --out_binsize 100 --hri --plot
    params      = Path(__file__).parents[1] / "tests" / "testparams" / "InterfParams.py"
    bed_path    = Path(__file__).parents[1] / "tests" / "testfiles" / "200kb_hri.csv"
    map_path    = Path(__file__).parents[1] / "tests" / "testfiles" / "200kb_hri.map"
    output_path = Path(__file__).parents[1] / "tests" / "testout" / "hri_200kb_test.B"

    cmd = BASE_CMD + [
        "--genome",
        "--params", str(params),
        "--bedgff", str(bed_path),
        "--rec_map", str(map_path),
        "--precise_chunks", "2",
        "--out", str(output_path),
        "--out_binsize", "100",
        "--hri",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    out = result.stdout + result.stderr

    # CLI ran
    assert result.returncode == 0, f"CLI failed:\n{result.stderr}"
    assert "Mean B of neutral sites across chromosome chr_200kb: 0.588623210410594" in out

    # File wrote
    assert output_path.exists(), "Expected output file not created"

    # Contents
    with open(output_path, "r") as f:
        lines = [ln.strip() for ln in f if ln.strip()]

    assert len(lines) == 2001, f"Expected 2001 lines, found {len(lines)}"

    # Specific rows must be present
    assert "chr_200kb,1,0.340332'" in lines
    assert "chr_200kb,101,0.340332'" in lines
    assert "chr_200kb,40101,0.359374'" in lines

    # Final line check
    assert lines[-1] == "chr_200kb,199901,0.327636'", f"Unexpected final line: {lines[-1]}"

